# Copyright 2018 Ian Campbell <ijc@hellion.org.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# Derived from the QLScrobbler plugin:
#
# QLScrobbler: an Audioscrobbler client plugin for Quod Libet.
# version 0.11
# (C) 2005-2023 by Joshua Kwan <joshk@triplehelix.org>,
#                  Joe Wreschnig <piman@sacredchao.net>,
#                  Franz Pletyz <fpletz@franz-pletz.org>,
#                  Nicholas J. Michalek <djphazer@gmail.com>,
#                  Steven Robertson <steven@strobe.cc>
#                  Nick Boultbee <nick.boultbee@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import os
import threading
import time

from gi.repository import Gtk, GLib

import quodlibet
from quodlibet import _
from quodlibet import app, qltk
from quodlibet.pattern import Pattern
from quodlibet.query import Query
from quodlibet.plugins.events import EventPlugin
from quodlibet.plugins import PluginConfig
from quodlibet.qltk.entry import ValidatingEntry, UndoEntry
from quodlibet.qltk.msg import Message
from quodlibet.qltk import Icons
from quodlibet.util.dprint import print_d
from quodlibet.util.picklehelper import pickle_load, pickle_dump, PickleError

import csv
from io import StringIO

from . import listenbrainz

DEFAULT_TITLEPAT = "<title><version| (<version>)>"
DEFAULT_ARTISTPAT = "<artist|<artist>|<composer|<composer>|<performer>>>"

plugin_config = PluginConfig("listenbrainz")
defaults = plugin_config.defaults
defaults.set("user_token", "")

defaults.set("titlepat", "")
defaults.set("artistpat", "")
defaults.set("exclude", "")
defaults.set("offline", False)
defaults.set("tags", "")


def config_get_title_pattern():
    return plugin_config.get("titlepat") or DEFAULT_TITLEPAT


def config_get_artist_pattern():
    return plugin_config.get("artistpat") or DEFAULT_ARTISTPAT


def config_get_tags():
    tags = plugin_config.get("tags") or None
    if tags is None:
        return []
    parser = csv.reader(StringIO(tags), quoting=csv.QUOTE_ALL, skipinitialspace=True)
    try:
        return next(parser)
    except Exception as e:
        print_d('Failed to parse tags "{}": {}'.format(*tags), e)
        return []


class ListenBrainzSubmitQueue:
    """Manages the submit queue for listens. Works independently of the
    plugin being enabled; other plugins may use submit() to queue songs for
    submission.
    """

    DUMP = os.path.join(quodlibet.get_user_dir(), "listenbrainz_cache")

    # These objects are shared across instances, to allow other plugins to
    # queue listens in future versions of QL.
    queue: list[tuple[int, listenbrainz.Track]] = []
    condition = threading.Condition()

    def set_nowplaying(self, song):
        """Send a Now Playing notification."""
        track = self._track(song)
        if not track or self.nowplaying_track == track:
            return
        print_d(f"Set now playing: {track}")
        self.condition.acquire()
        self.nowplaying_track = track
        self.nowplaying_sent = False
        self.changed()
        self.condition.release()

    def submit(self, song, timestamp=0):
        """Submit a song. If 'timestamp' is 0, the current time will
        be used."""
        track = self._track(song)
        if track is None:
            return

        self.condition.acquire()
        if timestamp == 0:
            timestamp = int(time.time())
        print_d(f"Queueing: {track}")
        self.queue.append((timestamp, track))
        self.changed()
        self.condition.release()

    def _track(self, song):
        """Returns a listenbrainz.Track."""
        artist = self.artpat.format(song)
        title = self.titlepat.format(song)
        if artist is None or title is None:
            return None

        album = song.comma("album")

        # https://listenbrainz.readthedocs.io/en/latest/dev/json.html#payload-json-details
        #
        # artist_mbids          A list of MusicBrainz Artist IDs, one or more Artist
        #                       IDs may be included here. If you have a complete
        #                       MusicBrainz artist credit that contains multiple
        #                       Artist IDs, include them all in this list.
        # release_group_mbid    A MusicBrainz Release Group ID of the release group this
        #                       recording was played from.
        # release_mbid          A MusicBrainz Release ID of the release this recording
        #                       was played from.
        # recording_mbid        A MusicBrainz Recording ID of the recording that was
        #                       played.
        # track_mbid            A MusicBrainz Track ID associated with the recording
        #                       that was played.
        # work_mbids            A list of MusicBrainz Work IDs that may be associated
        #                       with this recording.
        # tracknumber           The tracknumber of the recording. This first recording
        #                       on a release is tracknumber 1.
        # isrc                  The ISRC code associated with the recording.
        # spotify_id            The Spotify track URL associated with this recording.
        # tags                  A list of user defined tags to be associated with this
        #                       recording. These tags are similar to last.fm tags. For
        #                       example, you have apply tags such as punk, see-live,
        #                       smelly. You may submit up to MAX_TAGS_PER_LISTEN tags
        #                       and each tag may be up to MAX_TAG_SIZE characters large.
        #
        # https://picard.musicbrainz.org/docs/mappings/
        # Above			Tag
        # artists_mbids		MUSICBRAINZ_ARTISTID (multiple)
        # release_group_mbid	MUSICBRAINZ_RELEASEGROUPID
        # release_mbid		MUSICBRAINZ_ALBUMID
        # recording_mbid	MUSICBRAINZ_TRACKID
        # track_mbid		MUSICBRAINZ_RELEASETRACKID
        # work_mbids		MUSICBRAINZ_WORKID (multiple)
        # tracknumber		TRACKNUMBER
        # isrc			ISRC
        # spotify_id		N/A
        # tags			N/A
        additional_info = {}

        for k, v in [
            ("artist_mbids", song.list("musicbrainz_artistid")),
            ("release_group_mbid", song.get("musicbrainz_releasegroupid", None)),
            ("release_mbid", song.get("musicbrainz_albumid", None)),
            ("recording_mbid", song.get("musicbrainz_trackid", None)),
            ("track_mbid", song.get("musicbrainz_releasetrackid", None)),
            ("work_mbids", song.list("musicbrainz_workid")),
            ("tracknumber", song.get("tracknumber", None)),
            ("isrc", song.get("isrc", None)),
            ("tags", self.tags),
        ]:
            if v is not None and v != []:
                additional_info[k] = v

        print_d(f"Track({artist},{title},{album},{additional_info})")
        return listenbrainz.Track(artist, title, album, additional_info)

    def __init__(self):
        print("logging")
        self.nowplaying_track = None
        self.nowplaying_sent = False

        self.broken = False
        self.offline = False
        self.retries = 0

        self.lb = listenbrainz.ListenBrainzClient()  # XXX logger=xxx

        # These need to be set early for _format_song to work
        self.titlepat = Pattern(config_get_title_pattern())
        self.artpat = Pattern(config_get_artist_pattern())
        self.tags = config_get_tags()

        try:
            with open(self.DUMP, "rb") as disk_queue_file:
                disk_queue = pickle_load(disk_queue_file)
            os.unlink(self.DUMP)
            self.queue += disk_queue
        except (OSError, PickleError):
            pass

    @classmethod
    def dump_queue(cls):
        if cls.queue:
            try:
                with open(cls.DUMP, "wb") as disk_queue_file:
                    pickle_dump(cls.queue, disk_queue_file)
            except (OSError, PickleError):
                pass

    # Must be called with self.condition acquired
    def _check_config(self):
        user_token = plugin_config.get("user_token")
        if not user_token:
            if self.queue and not self.broken:
                self.quick_dialog(
                    _(
                        "Please visit the Plugins window to set "
                        "ListenBrainz up. Until then, listens will not be "
                        "submitted."
                    ),
                    Gtk.MessageType.INFO,
                )
                self.broken = True
        elif self.lb.user_token != user_token:
            print_d(f"Setting user_token {user_token}")
            self.lb.user_token = user_token
            self.broken = False
        self.offline = plugin_config.getboolean("offline")
        self.titlepat = Pattern(config_get_title_pattern())
        self.artpat = Pattern(config_get_artist_pattern())
        self.tags = config_get_tags()

    # Must be called with self.condition acquired
    def changed(self):
        """Signal that settings or queue contents were changed."""
        self._check_config()
        if (
            not self.broken
            and not self.offline
            and (self.queue or (self.nowplaying_track and not self.nowplaying_sent))
        ):
            self.condition.notify()

    def run(self):
        """Submit songs from the queue. Call from a daemon thread."""

        print_d("Submission queue thread running")

        while True:
            print_d("Top of queue loop")
            self.condition.acquire()

            while (
                self.broken
                or self.offline
                or (
                    not self.queue
                    and (not self.nowplaying_track or self.nowplaying_sent)
                )
            ):
                print_d("Nothing to do, waiting")
                self.condition.wait()
                print_d("Awoke")

            print_d("Running iteration")

            # Poll inputs under the lock

            submit = None
            if self.queue:
                submit = self.queue[0]
            nowplaying = None
            if self.nowplaying_track and not self.nowplaying_sent:
                nowplaying = self.nowplaying_track

            self.condition.release()

            # Call f() and handle errors with backoff and disable
            def with_backoff(f):
                try:
                    rsp = f()
                except Exception as e:
                    rsp = None
                    print_d(f"Error: {e}")

                if rsp and rsp.status == 200:
                    self.retries = 0
                    return True
                if self.retries >= 6:
                    # Too many retries, put self offline
                    print_d("Too many retries, setting to offline")
                    self.offline = True
                    plugin_config.set("offline", True)

                    self.quick_dialog(
                        _(
                            "Too many consecutive submission failures (%d). "
                            "Setting to offline mode. "
                            "Please visit the Plugins window to reset "
                            "ListenBrainz. Until then, listens will not be "
                            "submitted."
                        )
                        % self.retries,
                        Gtk.MessageType.INFO,
                    )
                    return False
                delay = 10
                print_d("Failure, waiting %ds" % delay)
                self.retries += 1
                time.sleep(delay)
                print_d("Done sleeping")
                return False
                return True

            if submit:
                (listened_at, track) = submit
                print_d(f"Submitting: {track}")

                if not with_backoff(lambda: self.lb.listen(listened_at, track)):  # noqa
                    continue

                print_d("Submission successful")

                # Remove submitted entry under lock
                self.condition.acquire()
                if self.queue[0] == submit:
                    self.queue.pop(0)
                self.condition.release()

            if nowplaying:
                print_d(f"Now playing: {nowplaying}")

                if not with_backoff(lambda: self.lb.playing_now(nowplaying)):  # noqa
                    continue

                print_d("Now playing submission successful")

                self.condition.acquire()
                if nowplaying == self.nowplaying_track:
                    # only if it didn't change under our feet
                    self.nowplaying_sent = True
                self.condition.release()

    def quick_dialog_helper(self, dialog_type, msg):
        dialog = Message(dialog_type, app.window, "ListenBrainz", msg)
        dialog.connect("response", lambda dia, resp: dia.destroy())
        dialog.show()

    def quick_dialog(self, msg, dialog_type):
        GLib.idle_add(self.quick_dialog_helper, dialog_type, msg)


class ListenbrainzSubmission(EventPlugin):
    PLUGIN_ID = "listenbrainz"
    PLUGIN_NAME = _("ListenBrainz Submission")
    PLUGIN_DESC = _("Submit listens to ListenBrainz.")
    PLUGIN_ICON = Icons.NETWORK_WORKGROUP

    def __init__(self):
        self.__enabled = False
        self.queue = ListenBrainzSubmitQueue()
        queue_thread = threading.Thread(None, self.queue.run)
        queue_thread.setDaemon(True)
        queue_thread.start()

        self.start_time = 0
        self.unpaused_time = 0
        self.elapsed = 0
        self.nowplaying = None

        self.exclude = plugin_config.get("exclude")

    def plugin_on_song_ended(self, song, stopped):
        if song is None or not self.__enabled:
            return
        if self.unpaused_time > 0:
            self.elapsed += time.time() - self.unpaused_time
        # https://listenbrainz.readthedocs.io/en/latest/dev/api.html
        #
        # Listens should be submitted for tracks when the user has
        # listened to half the track or 4 minutes of the track,
        # whichever is lower. If the user hasn’t listened to 4 minutes
        # or half the track, it doesn’t fully count as a listen and
        # should not be submitted.
        #
        # we check 'elapsed' rather than 'length' to work around wrong ~#length
        if self.elapsed < (4 * 60) and self.elapsed <= 0.5 * song.get("~#length", 0):
            return
        print_d(f"Checking against filter {self.exclude}")
        if self.exclude and Query(self.exclude).search(song):
            print_d("Not submitting: {}".format(song("~artist~title")))
            return
        self.queue.submit(song, self.start_time)

    def song_excluded(self, song):
        if self.exclude and Query(self.exclude).search(song):
            print_d(f"{song('~artist~title')} is excluded by {self.exclude}")
            return True
        return False

    def send_nowplaying(self, song):
        if not self.song_excluded(song):
            self.queue.set_nowplaying(song)

    def plugin_on_song_started(self, song):
        if song is None:
            return
        self.start_time = int(time.time())
        if app.player.paused:
            self.unpaused_time = 0
        else:
            self.unpaused_time = time.time()
        self.elapsed = 0
        if self.__enabled and not app.player.paused:
            self.send_nowplaying(song)
        else:
            self.nowplaying = song

    def plugin_on_paused(self):
        if self.unpaused_time > 0:
            self.elapsed += time.time() - self.unpaused_time
        self.unpaused_time = 0

    def plugin_on_unpaused(self):
        self.unpaused_time = time.time()
        if self.__enabled and self.nowplaying:
            self.send_nowplaying(self.nowplaying)
            self.nowplaying = None

    def enabled(self):
        self.__enabled = True
        print_d("Plugin enabled - accepting new songs.")

    def disabled(self):
        self.__enabled = False
        print_d("Plugin disabled - not accepting any new songs.")

    def PluginPreferences(self, parent):
        def changed(entry, key):
            if entry.get_property("sensitive"):
                plugin_config.set(key, entry.get_text())

        box = Gtk.VBox(spacing=12)

        # first frame
        table = Gtk.Table(n_rows=2, n_columns=2)
        table.props.expand = False
        table.set_col_spacings(6)
        table.set_row_spacings(6)

        labels = []
        label_names = [_("User _token:")]
        for idx, name in enumerate(label_names):
            label = Gtk.Label(label=name)
            label.set_alignment(0.0, 0.5)
            label.set_use_underline(True)
            table.attach(
                label,
                0,
                1,
                idx,
                idx + 1,
                xoptions=Gtk.AttachOptions.FILL | Gtk.AttachOptions.SHRINK,
            )
            labels.append(label)

        row = 0

        # token
        entry = UndoEntry()
        entry.set_text(plugin_config.get("user_token"))
        entry.connect("changed", changed, "user_token")
        table.attach(entry, 1, 2, row, row + 1)
        labels[row].set_mnemonic_widget(entry)
        row += 1

        box.pack_start(qltk.Frame(_("Account"), child=table), True, True, 0)

        # second frame
        table = Gtk.Table(n_rows=5, n_columns=2)
        table.props.expand = False
        table.set_col_spacings(6)
        table.set_row_spacings(6)

        label_names = [
            _("_Artist pattern:"),
            _("_Title pattern:"),
            _("T_ags:"),
            _("Exclude _filter:"),
        ]

        labels = []
        for idx, name in enumerate(label_names):
            label = Gtk.Label(label=name)
            label.set_alignment(0.0, 0.5)
            label.set_use_underline(True)
            table.attach(
                label,
                0,
                1,
                idx,
                idx + 1,
                xoptions=Gtk.AttachOptions.FILL | Gtk.AttachOptions.SHRINK,
            )
            labels.append(label)

        row = 0
        # artist pattern
        entry = UndoEntry()
        entry.set_text(plugin_config.get("artistpat"))
        entry.connect("changed", changed, "artistpat")
        table.attach(entry, 1, 2, row, row + 1)
        entry.set_tooltip_text(
            _(
                "The pattern used to format "
                "the artist name for submission. Leave blank for default."
            )
        )
        labels[row].set_mnemonic_widget(entry)
        row += 1

        # title pattern
        entry = UndoEntry()
        entry.set_text(plugin_config.get("titlepat"))
        entry.connect("changed", changed, "titlepat")
        table.attach(entry, 1, 2, row, row + 1)
        entry.set_tooltip_text(
            _(
                "The pattern used to format "
                "the title for submission. Leave blank for default."
            )
        )
        labels[row].set_mnemonic_widget(entry)
        row += 1

        # tags
        entry = UndoEntry()
        entry.set_text(plugin_config.get("tags"))
        entry.connect("changed", changed, "tags")
        table.attach(entry, 1, 2, row, row + 1)
        entry.set_tooltip_text(
            _(
                "List of tags to include in the submission. "
                "Comma-separated, use double-quotes if necessary."
            )
        )
        labels[row].set_mnemonic_widget(entry)
        row += 1

        # exclude filter
        entry = ValidatingEntry(Query.validator)
        entry.set_text(plugin_config.get("exclude"))
        entry.set_tooltip_text(_("Songs matching this filter will not be submitted."))
        entry.connect("changed", changed, "exclude")
        table.attach(entry, 1, 2, row, row + 1)
        labels[row].set_mnemonic_widget(entry)
        row += 1

        # offline mode
        offline = plugin_config.ConfigCheckButton(
            _("_Offline mode (don't submit anything)"), "offline", populate=True
        )
        table.attach(offline, 0, 2, row, row + 1)

        box.pack_start(qltk.Frame(_("Submission"), child=table), True, True, 0)

        return box



# Copyright 2006 Joe Wreschnig
#      2013-2022 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from gi.repository import Gtk

from quodlibet.formats import AudioFile
from quodlibet.plugins.gui import MenuItemPlugin
from quodlibet.plugins.songshelpers import is_a_file
from quodlibet.qltk.chooser import choose_folders
from quodlibet.qltk.download import DownloadProgress
from quodlibet.qltk.pluginwin import PluginWindow

from quodlibet import ngettext, _, print_d, app, util
from quodlibet import qltk
from quodlibet.errorreport import errorhook
from quodlibet.qltk.showfiles import show_songs

from quodlibet.util import print_e, print_w, copool
from quodlibet.qltk.msg import ConfirmationPrompt, ErrorMessage, Message
from quodlibet.qltk.delete import TrashMenuItem, trash_songs
from quodlibet.qltk.information import Information
from quodlibet.qltk.properties import SongProperties
from quodlibet.qltk.x import SeparatorMenuItem
from quodlibet.qltk.ratingsmenu import RatingsMenuItem
from quodlibet.qltk import get_top_parent, get_menu_item_top_parent, Icons
from quodlibet.plugins import PluginManager, PluginHandler
from quodlibet.plugins.songsmenu import SongsMenuPlugin
from quodlibet.util.songwrapper import list_wrapper, check_wrapper_changed


def confirm_song_removal_invoke(parent, songs):
    songs = set(songs)
    if not songs:
        return True

    count = len(songs)
    song = next(iter(songs))
    if count == 1:
        title = _('Remove track: "%(title)s" from the library?') % {
            "title": song("title") or song("~basename")
        }
    else:
        title = _("Remove %(count)d tracks from the library?") % {"count": count}

    prompt = ConfirmationPrompt(
        parent, title, "", _("Remove from Library"), ok_button_icon=Icons.LIST_REMOVE
    )
    return prompt.run() == ConfirmationPrompt.RESPONSE_INVOKE


def confirm_multi_song_invoke(parent, plugin_name, count):
    """Dialog to confirm invoking a plugin with X songs in case X is high"""
    title = ngettext(
        'Run the plugin "%(name)s" on %(count)d song?',
        'Run the plugin "%(name)s" on %(count)d songs?',
        count,
    ) % {"name": plugin_name, "count": count}
    description = ""
    ok_text = _("_Run Plugin")
    prompt = ConfirmationPrompt(parent, title, description, ok_text).run()
    return prompt == ConfirmationPrompt.RESPONSE_INVOKE


def confirm_multi_album_invoke(parent, plugin_name, count):
    """Dialog to confirm invoking a plugin with X albums in case X is high"""
    title = ngettext(
        'Run the plugin "%(name)s" on %(count)d album?',
        'Run the plugin "%(name)s" on %(count)d albums?',
        count,
    ) % {"name": plugin_name, "count": count}
    description = ""
    ok_text = _("_Run Plugin")
    prompt = ConfirmationPrompt(parent, title, description, ok_text).run()
    return prompt == ConfirmationPrompt.RESPONSE_INVOKE


class SongsMenuPluginHandler(PluginHandler):
    def __init__(self, song_confirmer=None, album_confirmer=None):
        """custom confirmers for testing"""

        self.__plugins = []

        self._confirm_multiple_songs = confirm_multi_song_invoke
        if song_confirmer is not None:
            self._confirm_multiple_songs = song_confirmer

        self._confirm_multiple_albums = confirm_multi_album_invoke
        if album_confirmer is not None:
            self._confirm_multiple_albums = album_confirmer

    def menu(self, library, songs):
        songs = list_wrapper(songs)

        attrs = ["plugin_song", "plugin_songs", "plugin_album", "plugin_albums"]

        if len(songs) == 1:
            attrs.append("plugin_single_song")

        last = (songs and songs[-1]) or None
        for song in songs:
            if song.album_key != last.album_key:
                break
            last = song
        else:
            attrs.append("plugin_single_album")

        items = []
        kinds = self.__plugins
        kinds.sort(key=lambda plugin: plugin.PLUGIN_ID)
        for Kind in kinds:
            usable = any(callable(getattr(Kind, s)) for s in attrs)
            if usable:
                try:
                    items.append(Kind(songs, library))
                except Exception:
                    print_e(f"Couldn't initialise song plugin {Kind}. Stack trace:")
                    errorhook()
        items = [i for i in items if i.initialized]

        if items:
            menu = Gtk.Menu()
            for item in items:
                try:
                    menu.append(item)
                    args = (library, songs)
                    if item.get_submenu():
                        for subitem in item.get_submenu().get_children():
                            subitem.connect("activate", self.__on_activate, item, *args)
                    else:
                        item.connect("activate", self.__on_activate, item, *args)
                except Exception:
                    errorhook()
                    item.destroy()
            menu.append(SeparatorMenuItem())
            prefs = Gtk.MenuItem(label=_("Configure Plugins…"))
            prefs.connect("activate", lambda _: PluginWindow().show())
            menu.append(prefs)

        else:
            menu = None
        return menu

    def __get_albums(self, songs):
        albums = {}
        for song in songs:
            key = song.album_key
            if key not in albums:
                albums[key] = []
            albums[key].append(song)

        albums = list(albums.values())
        for album in albums:
            album.sort()
        return albums

    def handle(self, plugin_id, library, parent, songs):
        """Start a song menu plugin directly without a menu"""

        parent = get_top_parent(parent)

        for plugin in self.__plugins:
            if plugin.PLUGIN_ID == plugin_id:
                songs = list_wrapper(songs)
                try:
                    plugin = plugin(songs, library)
                except Exception:
                    errorhook()
                else:
                    self.__handle(plugin, plugin, library, songs, parent)
                return

    def __on_activate(self, item, plugin, library, songs):
        parent = get_menu_item_top_parent(item)
        self.__handle(item, plugin, library, songs, parent)

    def __handle(self, item, plugin, library, songs, parent):
        if len(songs) == 0:
            return

        try:
            if len(songs) == 1 and callable(plugin.plugin_single_song):
                try:
                    ret = plugin.plugin_single_song(songs[0])
                except Exception:
                    errorhook()
                else:
                    if ret:
                        return
            if callable(plugin.plugin_song):
                total = len(songs)
                if total > plugin.MAX_INVOCATIONS:
                    if not self._confirm_multiple_songs(
                        parent, plugin.PLUGIN_NAME, total
                    ):
                        return

                try:
                    ret = map(plugin.plugin_song, songs)
                except Exception:
                    errorhook()
                else:
                    if any(ret):
                        return
            if callable(plugin.plugin_songs):
                try:
                    ret = plugin.plugin_songs(songs)
                except Exception:
                    errorhook()
                else:
                    if ret:
                        return

            if plugin.handles_albums:
                albums = self.__get_albums(songs)
                total = len(albums)
                if total > plugin.MAX_INVOCATIONS:
                    if not self._confirm_multiple_albums(
                        parent, plugin.PLUGIN_NAME, total
                    ):
                        return

            if callable(plugin.plugin_single_album) and len(albums) == 1:
                try:
                    ret = plugin.plugin_single_album(albums[0])
                except Exception:
                    errorhook()
                else:
                    if ret:
                        return
            if callable(plugin.plugin_album):
                try:
                    ret = map(plugin.plugin_album, albums)
                except Exception:
                    errorhook()
                else:
                    if any(ret):
                        return
            if callable(plugin.plugin_albums):
                try:
                    ret = plugin.plugin_albums(albums)
                except Exception:
                    errorhook()
                else:
                    if ret:
                        return

        finally:
            check_wrapper_changed(library, filter(None, songs))

    def plugin_handle(self, plugin):
        return issubclass(plugin.cls, SongsMenuPlugin)

    def plugin_enable(self, plugin):
        self.__plugins.append(plugin.cls)

    def plugin_disable(self, plugin):
        self.__plugins.remove(plugin.cls)


class SongsMenu(Gtk.Menu):
    plugins = SongsMenuPluginHandler()

    @classmethod
    def init_plugins(cls):
        PluginManager.instance.register_handler(cls.plugins)

    def __init__(
        self,
        library,
        songs,
        plugins=True,
        playlists=True,
        queue=True,
        remove=True,
        delete=False,
        edit=True,
        info=True,
        ratings=True,
        show_files=True,
        download=False,
        items=None,
        accels=True,
        removal_confirmer=None,
        folder_chooser=None,
    ):
        super().__init__()
        # The library may actually be a librarian; if it is, use it,
        # otherwise find the real librarian.
        librarian = getattr(library, "librarian", library)

        if ratings:
            ratings_item = RatingsMenuItem(songs, librarian)
            ratings_item.set_sensitive(bool(songs))
            self.append(ratings_item)
            self.separate()

        # external item groups
        for subitems in reversed(items or []):
            self.separate()
            for item in subitems:
                self.append(item)
            self.separate()

        if plugins:
            submenu = self.plugins.menu(librarian, songs)
            if submenu is not None:
                b = qltk.MenuItem(_("_Plugins"), Icons.SYSTEM_RUN)
                b.set_sensitive(bool(songs))
                self.append(b)
                b.set_submenu(submenu)
                self.append(SeparatorMenuItem())

        in_lib = True
        can_add = True
        is_file = True
        for song in songs:
            if song not in library:
                in_lib = False
            if not song.can_add:
                can_add = False
            if not song.is_file:
                is_file = False

        if playlists:
            self.init_playlists(can_add, library, songs)
        if queue:
            self.init_queue(accels, can_add, songs)

        if remove or delete:
            self.separate()

        if remove:
            self.init_remove(in_lib, library, removal_confirmer, remove, songs)

        if delete:
            self.init_delete(accels, delete, is_file, songs, librarian)

        if edit:
            self.init_edit(accels, songs, librarian)

        if info:
            self.init_info(accels, songs, librarian)

        if show_files and any(is_a_file(s) for s in songs):
            self.init_show_files(songs)

        if download:
            self.init_download(songs, folder_chooser)

        def selection_done_cb(menu):
            menu.destroy()

        self.connect("selection-done", selection_done_cb)

    def init_download(self, songs, folder_chooser):
        def is_downloadable(song: AudioFile):
            return bool(not song.is_file and song.get("~uri", False))

        self.separate()
        relevant = [s for s in songs if is_downloadable(s)]
        total = len(relevant)
        text = ngettext("_Download file…", "_Download %(total)d files…", total) % {
            "total": total
        }
        b = qltk.MenuItem(text, Icons.EMBLEM_DOWNLOADS)
        b.set_sensitive(relevant and len(relevant) < MenuItemPlugin.MAX_INVOCATIONS)

        def _finished(p, successes, failures):
            msg = (
                f"{util.bold(successes)} "
                + _("successful")
                + f"\n{util.bold(failures)} "
                + _("failed")
            )
            print_d(msg.replace("\n", "; "))
            warning = Message(
                Gtk.MessageType.INFO,
                app.window,
                _("Downloads complete"),
                msg,
                escape_desc=False,
            )
            warning.run()

        def download_cb(menu_item):
            songs = relevant
            total = len(songs)
            msg = ngettext("Download {name!r} to", "Download {total} files to", total)
            msg = msg.format(
                name=next(iter(songs))("title")[:99] if total else "?", total=total
            )
            chooser = folder_chooser or choose_folders
            paths = chooser(None, msg, _("Download here"), allow_multiple=False)
            if not paths:
                print_d("Cancelling download")
                return
            path = paths[0]
            progress = DownloadProgress(songs)

            progress.connect("finished", _finished)
            copool.add(progress.download_songs, path)

        b.connect("activate", download_cb)
        self.append(b)

    def init_show_files(self, songs):
        def show_files_cb(menu_item):
            print_d("Trying to show files...")
            if not show_songs(songs):
                parent = get_menu_item_top_parent(menu_item)
                msg = ErrorMessage(
                    parent,
                    _("Unable to show files"),
                    _("Error showing files, or no program available to show them."),
                )
                msg.run()

        self.separate()
        total = len([s for s in songs if is_a_file(s)])
        text = ngettext(
            "_Show in File Manager", "_Show %(total)d Files in File Manager", total
        ) % {"total": total}
        b = qltk.MenuItem(text, Icons.DOCUMENT_OPEN)
        b.set_sensitive(bool(songs) and len(songs) < MenuItemPlugin.MAX_INVOCATIONS)
        b.connect("activate", show_files_cb)
        self.append(b)

    def init_info(self, accels, songs, librarian):
        b = qltk.MenuItem(_("_Information"), Icons.DIALOG_INFORMATION)
        b.set_sensitive(bool(songs))
        if accels:
            qltk.add_fake_accel(b, "<Primary>I")

        def information_cb(menu_item):
            parent = get_menu_item_top_parent(menu_item)
            window = Information(librarian, songs, parent)
            window.show()

        b.connect("activate", information_cb)
        self.append(b)

    def init_edit(self, accels, songs, librarian):
        self.separate()
        b = qltk.MenuItem(_("Edit _Tags"), Icons.EDIT)
        b.set_sensitive(bool(songs))
        if accels:
            qltk.add_fake_accel(b, "<alt>Return")

        def song_properties_cb(menu_item):
            parent = get_menu_item_top_parent(menu_item)
            window = SongProperties(librarian, songs, parent)
            window.show()

        b.connect("activate", song_properties_cb)
        self.append(b)

    def init_delete(self, accels, delete, is_file, songs, librarian):
        if callable(delete):
            b = qltk.MenuItem(_("_Delete"), Icons.EDIT_DELETE)
            b.connect("activate", lambda item: delete(songs))
            if accels:
                qltk.add_fake_accel(b, "<Primary>Delete")
        else:
            b = TrashMenuItem()
            if accels:
                qltk.add_fake_accel(b, "<Primary>Delete")

            def trash_cb(item):
                parent = get_menu_item_top_parent(item)
                trash_songs(parent, songs, librarian)

            b.connect("activate", trash_cb)
            b.set_sensitive(is_file and bool(songs))
        self.append(b)

    def init_remove(self, in_lib, library, removal_confirmer, remove, songs):
        self._confirm_song_removal = removal_confirmer or confirm_song_removal_invoke
        b = qltk.MenuItem(_("_Remove from Library…"), Icons.LIST_REMOVE)
        if callable(remove):
            b.connect("activate", lambda item: remove(songs))
        else:

            def remove_cb(item, songs, library):
                parent = get_menu_item_top_parent(item)
                if self._confirm_song_removal(parent, songs):
                    library.remove(songs)

            b.connect("activate", remove_cb, songs, library)
            b.set_sensitive(in_lib and bool(songs))
        self.append(b)

    def init_queue(self, accels, can_add, songs):
        b = qltk.MenuItem(_("Add to _Queue"), Icons.LIST_ADD)

        def enqueue_cb(item, songs):
            songs = [s for s in songs if s.can_add]
            if songs:
                from quodlibet import app

                app.window.playlist.enqueue(songs)

        b.connect("activate", enqueue_cb, songs)
        if accels:
            qltk.add_fake_accel(b, "<Primary>Return")
        self.append(b)
        b.set_sensitive(can_add and bool(songs))

    def init_playlists(self, can_add, library, songs):
        try:
            from quodlibet.browsers.playlists.menu import PlaylistMenu

            submenu = PlaylistMenu(songs, library.playlists)
        except AttributeError as e:
            print_w(f"Couldn't get Playlists menu: {e}")
        else:
            b = qltk.MenuItem(_("Play_lists"), Icons.FOLDER_DRAG_ACCEPT)
            b.set_sensitive(can_add and bool(songs))
            b.set_submenu(submenu)
            self.append(b)

    def separate(self):
        if not self.get_children():
            return
        if not isinstance(self.get_children()[-1], Gtk.SeparatorMenuItem):
            self.append(SeparatorMenuItem())

    def preseparate(self):
        if not self.get_children():
            return
        if not isinstance(self.get_children()[0], Gtk.SeparatorMenuItem):
            self.prepend(SeparatorMenuItem())



# Copyright 2004-2013 Joe Wreschnig, Michael Urman, Iñigo Serna,
#                     Christoph Reiter, Steven Robertson
#           2011-2024 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from __future__ import annotations

import os
import random
from typing import Any
from urllib.parse import quote

from senf import fsnative, fsn2bytes, bytes2fsn, path2fsn, _fsnative, uri2fsn, fsn2uri

from quodlibet import ngettext, _
from quodlibet import util
from quodlibet import config
from quodlibet.formats._audio import (
    TAG_TO_SORT,
    NUMERIC_ZERO_DEFAULT,
    AudioFile,
    HasKey,
)
from quodlibet.formats._audio import PEOPLE as _PEOPLE
from quodlibet.pattern import Pattern
from quodlibet.const import QL_NAMESPACE

try:
    from collections import abc
except ImportError:
    import collections as abc  # type: ignore

from quodlibet.util import is_windows
from quodlibet.util.path import escape_filename, unescape_filename, limit_path
from quodlibet.util.dprint import print_d, print_w
from quodlibet.util.misc import total_ordering, hashable
from .collections import HashedList
from datetime import datetime
from os.path import splitext, basename, dirname, exists
from xml.etree import ElementTree as ET
from xml.etree.ElementTree import ElementTree, Element

XSPF_NS = "http://xspf.org/ns/0/"

PEOPLE = list(_PEOPLE)
# Collections value albumartist more than song artist (Issue 1034)
PEOPLE.remove("albumartist")
PEOPLE.insert(0, "albumartist")

ELPOEP = list(reversed(PEOPLE))
PEOPLE_SCORE = [100**i for i in range(len(PEOPLE))]


def avg(nums):
    """Returns the average (arithmetic mean) of a list of numbers"""
    return float(sum(nums)) / len(nums)


def bayesian_average(nums, c=None, m=None):
    """Returns the Bayesian average of an iterable of numbers,
    with parameters defaulting to config specific to ~#rating."""
    m = m or config.RATINGS.default
    c = c or config.getfloat("settings", "bayesian_rating_factor", 0.0)
    return float(m * c + sum(nums)) / (c + len(nums))


NUM_DEFAULT_FUNCS = {
    "length": "sum",
    "playcount": "sum",
    "added": "max",
    "lastplayed": "max",
    "laststarted": "max",
    "mtime": "max",
    "rating": "bav",
    "skipcount": "sum",
    "year": "min",
    "originalyear": "min",
    "filesize": "sum",
}

NUM_FUNCS = {"max": max, "min": min, "sum": sum, "avg": avg, "bav": bayesian_average}


class Collection:
    """A collection of songs which implements some methods similar to the
    AudioFile class.

    The content of the collection can be changed by changing the content of
    the songs attribute.
    """

    _cache_size = 6
    songs = ()

    def __init__(self):
        """Cache in _cache, LRU key order in _used, keys that return default
        are in _default"""
        self.__cache = {}
        self.__default = set()
        self.__used = []

    def finalize(self):
        """Finalize the collection.
        Call this after songs get added or removed"""
        self.__cache.clear()
        self.__default.clear()
        self.__used = []

    def get(self, key, default="", connector=" - "):
        if not self.songs:
            return default
        if key[:1] == "~" and "~" in key[1:]:
            if not isinstance(default, str):
                return default
            keys = util.tagsplit(key)
            v = map(self.__get_cached_value, keys)

            def default_funct(x):
                if x is None:
                    return default
                return x

            v = map(default_funct, v)
            v = ((isinstance(x, float) and f"{x:.2f}") or x for x in v)
            v = (isinstance(x, str) and x or str(x) for x in v)
            return connector.join(filter(None, v)) or default
        value = self.__get_cached_value(key)
        if value is None:
            return default
        return value

    __call__ = get

    def comma(self, key):
        value = self.get(key)
        return value if isinstance(value, int | float) else value.replace("\n", ", ")

    def list(self, key):
        v = self.get(key, connector="\n") if "~" in key[1:] else self.get(key)
        if isinstance(v, float):
            # Ignore insignificant differences in numeric tags caused
            # by floating point imprecision when converting them to strings
            v = round(v, 8)
        return [] if v == "" else str(v).split("\n")

    def __get_cached_value(self, key):
        if key in self.__cache:
            self.__used.remove(key)
            self.__used.insert(0, key)
            return self.__cache[key]
        if key in self.__default:
            return None
        val = self.__get_value(key)
        if val is None:
            self.__default.add(key)
        else:
            self.__used.insert(0, key)
            self.__cache[key] = val
        # Remove the oldest if the cache is full
        if len(self.__used) > self._cache_size:
            self.__cache.pop(self.__used.pop(-1))
        return val

    def __get_value(self, key):
        """This is similar to __call__ in the AudioFile class.
        All internal tags are changed to represent a collection of songs.
        """

        # Using key:<func> runs the resulting list of values
        # through the function before returning it.
        # Numeric keys without a func will default to a reasonable function
        if key.startswith("~#"):
            key = key[2:]

            if key[-4:-3] == ":":
                func = key[-3:]
                key = key[:-4]
            elif key == "tracks":
                return len(self.songs)
            elif key == "discs":
                return len({song("~#disc", 1) for song in self.songs})
            elif key == "bitrate":
                length = self.__get_value("~#length")
                if not length:
                    return 0

                def w(s):
                    return s("~#bitrate", 0) * s("~#length", 0)

                return sum(w(song) for song in self.songs) / length
            else:
                # Standard or unknown numeric key.
                # AudioFile will try to cast the values to int,
                # default to avg
                func = NUM_DEFAULT_FUNCS.get(key, "avg")

            key = "~#" + key
            func = NUM_FUNCS.get(func)
            if func:
                # If none of the songs can return a numeric key,
                # the album returns default
                values = (song(key) for song in self.songs)
                values = [v for v in values if v != ""]
                return func(values) if values else None
            if key in NUMERIC_ZERO_DEFAULT:
                return 0
            return None
        if key[:1] == "~":
            key = key[1:]
            numkey = key.split(":")[0]
            keys = {"people": {}, "peoplesort": {}}
            if key in keys:
                people = keys["people"]
                peoplesort = keys["peoplesort"]
                for song in self.songs:
                    # Rank people by "relevance" -- artists before composers
                    # before performers, then by number of appearances.
                    for w, k in enumerate(ELPOEP):
                        persons = song.list(k)
                        for person in persons:
                            people[person] = people.get(person, 0) - PEOPLE_SCORE[w]
                        if k in TAG_TO_SORT:
                            persons = song.list(TAG_TO_SORT[k]) or persons
                        for person in persons:
                            peoplesort[person] = (
                                peoplesort.get(person, 0) - PEOPLE_SCORE[w]
                            )
                # It's cheaper to get people and peoplesort in one go
                keys["people"] = sorted(people.keys(), key=people.__getitem__)[:100]
                keys["peoplesort"] = sorted(
                    peoplesort.keys(), key=peoplesort.__getitem__
                )[:100]

                ret = keys.pop(key)
                ret = (ret and "\n".join(ret)) or None

                other, values = keys.popitem()
                other = "~" + other
                if not values:
                    self.__default.add(other)
                else:
                    if other in self.__used:
                        self.__used.remove(other)
                    self.__used.append(other)
                    self.__cache[other] = "\n".join(values)
                return ret
            if numkey == "length":
                length = self.__get_value("~#" + key)
                return None if length is None else util.format_time(length)
            if numkey == "long-length":
                length = self.__get_value("~#" + key[5:])
                return None if length is None else util.format_time_long(length)
            if numkey == "tracks":
                tracks = self.__get_value("~#" + key)
                return (
                    None
                    if tracks is None
                    else ngettext("%d track", "%d tracks", tracks) % tracks
                )
            if numkey == "discs":
                discs = self.__get_value("~#" + key)
                if discs > 1:
                    return ngettext("%d disc", "%d discs", discs) % discs
                # TODO: check this is correct for discs == 1
                return None
            if numkey == "rating":
                rating = self.__get_value("~#" + key)
                if rating is None:
                    return None
                return util.format_rating(rating)
            if numkey == "filesize":
                size = self.__get_value("~#" + key)
                return None if size is None else util.format_size(size)
            key = "~" + key

        # Nothing special was found, so just take all values of the songs
        # and sort them by their number of appearance
        result = {}
        for song in self.songs:
            for value in song.list(key):
                result[value] = result.get(value, 0) - 1

        values = [x[0] for x in sorted(result.items(), key=lambda x: (x[1], x[0]))]
        return "\n".join(values) if values else None


class Album(Collection, HasKey):
    """Like a `Collection` but adds cover scanning, some attributes for sorting
    and uses a set for the songs."""

    @util.cached_property
    def peoplesort(self):
        return util.human_sort_key(self.get("~peoplesort").split("\n")[0])

    @util.cached_property
    def genre(self):
        return util.human_sort_key(self.get("genre").split("\n")[0])

    @property
    def date(self):
        return self.get("date")

    @property
    def title(self):
        return self.get("album")

    def __init__(self, song):
        super().__init__()
        self.songs = set()
        # albumsort is part of the album_key, so every song has the same
        self.sort = util.human_sort_key(song("albumsort"))
        self.key = song.album_key

    @property
    def str_key(self):
        return str(self.key)

    def finalize(self):
        """Finalize this album. Call after songs get added or removed"""
        super().finalize()
        self.__dict__.pop("peoplesort", None)
        self.__dict__.pop("genre", None)

    def __repr__(self):
        return f"Album({self.key!r})"


@hashable
@total_ordering
class Playlist(Collection, abc.Iterable, HasKey):
    """A Playlist is a `Collection` that has list-like features
    Songs can appear more than once.
    """

    @staticmethod
    def suggested_name_for(songs):
        if len(songs) == 0:
            return _("Empty Playlist")
        if len(songs) == 1:
            return songs[0].comma("title")
        return ngettext(
            "%(title)s and %(count)d more",
            "%(title)s and %(count)d more",
            len(songs) - 1,
        ) % ({"title": songs[0].comma("title"), "count": len(songs) - 1})

    def __init__(self, name: str, songs_lib=None, pl_lib=None):
        super().__init__()
        self._list: HashedList = HashedList()
        # we require a file library here with masking
        assert songs_lib is None or hasattr(songs_lib, "masked")
        self.songs_lib = songs_lib

        name = str(name)
        if not name:
            raise ValueError("Playlists must have a name")
        self.name = name

        self.pl_lib = pl_lib
        # Libraries are dict-like so falsey if empty
        if self.pl_lib is None:
            print_w(f"Playlist {name!r} initialised without library")
        else:
            self.pl_lib.add([self])
        self.__inhibit_library_signals = False

    @property
    def key(self) -> str:  # type: ignore  # (Note: we want no setter)
        return self.name

    def get(self, key, default="", connector=" - "):
        if key == "~name":
            return self.name
        return super().get(key, default, connector)

    __call__ = get

    # List-like methods, for compatibility with original Playlist class.
    def extend(self, songs: abc.Iterable[AudioFile]):
        self._list.extend(songs)
        self.finalize()
        self._emit_changed(songs, msg="extend")

    def append(self, song):
        ret = self._list.append(song)
        self._emit_changed([song], msg="append")
        self.finalize()
        return ret

    def clear(self):
        self._emit_changed(self._list, msg="clear")
        del self._list[:]
        self.finalize()

    def __iter__(self):
        return iter(self._list)

    def __len__(self):
        return len(self._list)

    def __getitem__(self, index):
        return self._list[index]

    def index(self, value):
        return self._list.index(value)

    def __setitem__(self, key, value):
        self._list[key] = value
        if isinstance(key, slice):
            # TODO: more reliable slice support
            self._emit_changed(value, msg="direct slice set")
        else:
            self._emit_changed([value], msg="direct set")
        self.finalize()

    @property
    def songs(self):
        return [s for s in self._list if not isinstance(s, str)]

    def rename(self, new_name):
        """Changes this playlist's name and re-saves, or raises an `ValueError`
        if the name is not allowed"""
        old_name = self.key
        if new_name == old_name:
            return
        self.name = self._validated_name(new_name)
        self.write()
        if self.pl_lib:
            self.pl_lib.rename(self, old_name)

    def _validated_name(self, new_name):
        """Returns a transformed (or not) name, or raises a `ValueError`
        if the name is not allowed
        """

        new_name = str(new_name)
        if not new_name:
            raise ValueError(_("Playlists must have a name"))
        return new_name

    def add_songs(self, filenames, library):
        # TODO: use Set here to avoid duplicate signals
        changed = []
        for i in range(len(self)):
            if isinstance(self[i], str) and self._list[i] in filenames:
                song = library[self._list[i]]
                self._list[i] = song
                changed.append(song)
        if changed:
            self._emit_changed(changed, msg="add")
        return bool(changed)

    def remove_songs(
        self, songs: abc.Iterable[AudioFile], leave_dupes: bool = False
    ) -> bool:
        """Removes `songs` from this playlist if they are there,
        removing only the first reference if `leave_dupes` is True
        :returns True if anything was removed
        """
        changed = False
        for song in songs:
            # TODO: document the "library.masked" business
            if self.songs_lib is not None and self.songs_lib.masked(song):
                while True:
                    try:
                        self._list[self.index(song)] = song("~filename")
                    except ValueError:
                        break
                    else:
                        changed = True
            else:
                while song in self._list:
                    self._list.remove(song)
                    changed = True
                    if leave_dupes:
                        break

        def songs_gone():
            return set(songs) - set(self._list)

        if changed:
            self.finalize()
            # Short-circuit logic will avoid the calculation
            if not leave_dupes or songs_gone():
                self._emit_changed(songs, "remove_songs")
        return changed

    @property
    def inhibit(self):
        return self.__inhibit_library_signals

    @inhibit.setter
    def inhibit(self, value):
        self.__inhibit_library_signals = value

    def _emit_changed(self, songs, msg=""):
        if self.pl_lib is not None and not self.inhibit:
            print_d(f"Changed playlist {self!r} ({msg})")
            self.inhibit = True
            # Awkward, but we don't want to make Collection a GObject really
            self.pl_lib.emit("changed", [self])
            self.inhibit = False
        if self.songs_lib and not self.inhibit and songs:
            # See above re: emitting
            self.songs_lib.emit("changed", songs)

    def has_songs(self, songs):
        # TODO(rm): consider the "library.masked" business
        some, all = False, True
        for song in songs:
            found = song in self._list
            some = some or found
            all = all and found
            if some and not all:
                break
        return some, all

    def delete(self):
        if self.pl_lib is not None and not self.inhibit:
            self.pl_lib.remove([self])

    def write(self):
        pass

    @property
    def has_duplicates(self):
        """Returns True if there are any duplicated files in this playlist"""
        return self._list.has_duplicates()

    def shuffle(self):
        """Randomly shuffles this playlist, without weighting"""
        random.shuffle(self._list)
        if self.pl_lib is not None:
            self.pl_lib.changed([self])
        self.write()

    def __eq__(self, other):
        try:
            return self.name == other.name
        except AttributeError:
            return False

    def __lt__(self, other):
        try:
            return self.name < other.name
        except AttributeError:
            return False

    def __hash__(self):
        return id(self)

    def __str__(self):
        songs_text = ngettext("%d song", "%d songs", len(self.songs)) % len(self.songs)
        return f'"{self.name}" ({songs_text})'


class FileBackedPlaylist(Playlist):
    """A `Playlist` that is stored as a UTF-8 text file of paths"""

    def __init__(
        self,
        dir_: _fsnative,
        filename: _fsnative,
        songs_lib=None,
        pl_lib=None,
        validate: bool = False,
    ):
        assert isinstance(dir_, fsnative)
        self.dir = dir_
        name = self.name_for(filename)
        super().__init__(name, songs_lib=songs_lib, pl_lib=pl_lib)
        if validate:
            self.name = self._validated_name(name)
        # Store the actual filename used, not sanitised and validated name
        # This means we can delete imported things properly, etc...
        self._last_fn = os.path.join(dir_, filename)
        try:
            self._populate_from_file()
        except OSError:
            if self.name:
                print_d(f"Playlist '{self.name}' not found, creating new.")
                self.write()

    @classmethod
    def name_for(cls, filename: _fsnative) -> str:
        return unescape_filename(filename)

    @classmethod
    def filename_for(cls, filename: str):
        return escape_filename(filename)

    def _populate_from_file(self):
        """Populates, or raises IOError if no file found"""
        library = self.songs_lib
        with open(self.path, "rb") as h:
            for line in h:
                assert library is not None
                try:
                    line = bytes2fsn(line.rstrip(), "utf-8")
                except ValueError:
                    # decoding failed
                    continue
                if line in library:
                    self._list.append(library[line])
                elif library and library.masked(line):
                    self._list.append(line)

    @classmethod
    def new(cls, dir_: _fsnative, base: str | None = None, songs_lib=None, pl_lib=None):
        base = base or _("New Playlist")
        assert isinstance(dir_, fsnative)

        if not (dir_ and os.path.realpath(dir_)):
            raise ValueError(f"Invalid playlist directory {dir_!r}")

        last_error = None
        for i in range(1000):
            name = "%s %d" % (base, i) if i else base
            fn = cls.filename_for(name)
            try:
                return cls(dir_, fn, songs_lib=songs_lib, pl_lib=pl_lib, validate=True)
            except ValueError as e:
                last_error = e
        raise ValueError(
            f"Couldn't create playlist of name '{base}' (e.g. {last_error})"
        )

    @classmethod
    def from_songs(cls, dir_, songs, title=None, songs_lib=None, pl_lib=None):
        assert isinstance(dir_, fsnative)
        title = title or cls.suggested_name_for(songs)
        playlist = cls.new(dir_, title, songs_lib=songs_lib, pl_lib=pl_lib)
        playlist.extend(songs)
        return playlist

    @property
    def path(self):
        return os.path.join(self.dir, self.filename_for(self.name))

    @property
    def key(self) -> str:  # type: ignore  # (Note: we want no setter)
        return self.name

    def _validated_name(self, new_name: str) -> str:
        new_name = super()._validated_name(new_name)
        path = os.path.join(self.dir, self.filename_for(new_name))
        if os.path.exists(path):
            raise ValueError(
                _("A playlist named %(name)s already exists at %(path)s")
                % {"name": new_name, "path": path}
            )
        return new_name

    def delete(self):
        self._delete_file(self._last_fn)
        super().delete()

    @classmethod
    def _delete_file(cls, fn):
        print_d(f"Deleting playlist file: {fn!r}")
        try:
            os.unlink(fn)
        except OSError as e:
            print_w(f"Couldn't delete {fn!r} ({e})")

    def write(self):
        fn = self.path
        with open(fn, "wb") as f:
            for song in self._list:
                if isinstance(song, str):
                    f.write(fsn2bytes(song, "utf-8") + b"\n")
                else:
                    f.write(fsn2bytes(song("~filename"), "utf-8") + b"\n")
        if self._last_fn != fn:
            self._delete_file(self._last_fn)
            self._last_fn = fn

    def __str__(self):
        return f"<{type(self).__name__} at {self.path!r}>"


class XSPFBackedPlaylist(FileBackedPlaylist):
    _VERSION: int = 2
    """Persistence version"""

    EXT = "xspf"
    CREATOR_PATTERN = Pattern("<artist|<artist>|<~people>>")
    _SAFER = {c: quote(c, safe="") for c in ('\\/:*?"<>|' if is_windows() else "\0/")}

    @classmethod
    def from_playlist(cls, old_pl: FileBackedPlaylist, songs_lib, pl_lib):
        """Migrate from an existing file-based playlist"""

        def backup_for(path: str) -> str:
            base = os.path.join(dirname(path), ".backup")
            if not exists(base):
                print_d(f"Creating playlist backup directory {base}")
                os.mkdir(base)
            return os.path.join(base, basename(path))

        name = old_pl.name
        new = XSPFBackedPlaylist.new(
            old_pl.dir, name, songs_lib=songs_lib, pl_lib=pl_lib
        )
        new.extend(old_pl)
        new.write()
        os.rename(old_pl.path, backup_for(old_pl.path))
        old_pl.delete()
        return new

    def _populate_from_file(self):
        library = self.songs_lib
        try:
            tree = ET.parse(self.path)
            # TODO: validate some top-level tag data

            root = tree.getroot()
            if root.tag == "playlist":
                ns_mapping = None
                print_w(f"Using legacy namespace for import of {self.path}")
            elif root.tag == "{" + XSPF_NS + "}playlist":
                # Try correct format first
                ns_mapping = {"": XSPF_NS}
            else:
                raise ValueError(f"Unknown playlist root of {root.tag}")
            node = root.find("title", namespaces=ns_mapping)
            if node is None:
                print_w(
                    f"No <title> found in {self.path} "
                    f"(Got nodes: {[el.tag for el in root.iter()]})"
                )
            elif self.name != node.text:
                print_w(
                    f"Playlist was named {node.text!r} in XML "
                    f"instead of {self.name!r} at {self.path!r}"
                )

            for node in tree.iterfind(".//track", namespaces=ns_mapping):
                location = node.findtext("location", namespaces=ns_mapping).strip()
                path = location.replace("\n", "").replace("\r", "")
                try:
                    # TODO: process relative URIs too?
                    path = uri2fsn(path)
                except ValueError:
                    pass
                if path in library:
                    self._list.append(library[path])
                elif library and library.masked(path):
                    self._list.append(path)
                else:
                    # TODO: handle missing playlist items (#3105, #729, #3131)
                    node_dump = ET.tostring(node, method="xml").decode("utf-8")
                    print_w(
                        f"Couldn't find {path!r} in playlist at {self.path!r}. "
                        f"Perhaps its metadata will help: {node_dump!r}"
                    )
                    self._list.append(path)
                    library.mask(path)
        except (ET.ParseError, ValueError) as e:
            print_w(f"Couldn't load {self.path!r} ({e})")

    @classmethod
    def filename_for(cls, name: str):
        # Manually do *minimal* escaping, to allow near-readable filenames
        for bad, good in cls._SAFER.items():
            name = name.replace(bad, good)
        return path2fsn(f"{limit_path(name)}.{cls.EXT}")

    @classmethod
    def name_for(cls, file_path: _fsnative) -> str:
        filename, ext = splitext(unescape_filename(file_path))
        if not ext or ext.lower() != f".{cls.EXT}":
            raise TypeError(f"XSPFs should end in '{cls.EXT}', not {ext}")
        return filename

    def write(self):
        track_list = Element("trackList")
        # TODO: ditch for proper indent, once we have Python 3.9
        track_list.text = "\n"
        for song in self._list:
            if isinstance(song, str):
                track = {"location": fsn2uri(song)}
            else:
                creator = self.CREATOR_PATTERN.format(song)
                mbid = song("musicbrainz_trackid")
                track = {
                    "location": song("~uri"),
                    "identifier": (
                        f"https://musicbrainz.org/recording/{mbid}" if mbid else None
                    ),
                    "title": song("title"),
                    "creator": creator,
                    "album": song("album"),
                    "trackNum": song("~#track"),
                    "duration": int(song("~#length") * 1000.0),
                }
            track_list.append(self._element_from("track", track, True))
        playlist = Element("playlist", attrib={"version": "1", "xmlns": XSPF_NS})
        # Be kind to cat, git, editors etc. by leaving a final newline
        playlist.tail = "\n"
        playlist.append(self._version_tag())
        playlist.append(self._text_element("title", self.name))
        playlist.append(self._text_element("date", datetime.now().isoformat()))
        playlist.append(track_list)
        tree = ElementTree(playlist)
        ET.register_namespace("", XSPF_NS)
        path = self.path
        print_d(f"Writing {path !r}")
        tree.write(path, encoding="utf-8", xml_declaration=True)
        if self._last_fn != path:
            self._delete_file(self._last_fn)
            self._last_fn = path

    @classmethod
    def _version_tag(cls):
        """Currently this isn't needed (version 2 is detected),
        but it probably will be and easier to know than to guess"""
        meta = Element("meta", attrib={"rel": f"{QL_NAMESPACE}/playlists/version"})
        meta.text = str(cls._VERSION)
        return meta

    @classmethod
    def _text_element(cls, name: str, value: Any) -> Element:
        el = Element(name)
        el.text = str(value)
        return el

    @classmethod
    def _element_from(cls, name: str, d: dict, with_newline: bool = False) -> Element:
        """Converts a dict to XML etree. Removes falsey nodes"""
        out = Element(name)
        for k, v in d.items():
            if k and v:
                element = (
                    cls._element_from(k, v)
                    if isinstance(v, dict)
                    else cls._text_element(k, v)
                )
                out.append(element)
        if with_newline:
            out.tail = "\n"
        return out



# Copyright 2014-2024 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import socket
import time
from urllib.parse import quote, unquote

from quodlibet import _
from quodlibet import app
from quodlibet.util.dprint import print_w, print_d, print_


class SqueezeboxError(Exception):
    """Errors communicating with the Squeezebox"""


class SqueezeboxServerSettings(dict):
    """Encapsulates Server settings"""

    def __str__(self):
        try:
            return _("Squeezebox server at {hostname}:{port}").format(**self)
        except KeyError:
            return _("unidentified Squeezebox server")


class SqueezeboxPlayerSettings(dict):
    """Encapsulates player settings"""

    def __str__(self):
        try:
            return "{name} [{playerid}]".format(**self)
        except KeyError:
            return _("unidentified Squeezebox player: %r") % self


class SqueezeboxServer:
    """Encapsulates access to a Squeezebox player via a squeezecenter server"""

    _TIMEOUT = 10
    _MAX_FAILURES = 3
    telnet = None
    is_connected = False
    current_player = 0
    players: list[SqueezeboxPlayerSettings] = []
    config = SqueezeboxServerSettings()
    _debug = False

    def __init__(
        self,
        hostname="localhost",
        port=9090,
        user="",
        password="",
        library_dir="",
        current_player=0,
        debug=False,
    ):
        self._debug = debug
        self.failures = 0
        self.delta = 600  # Default in ms
        self.config = SqueezeboxServerSettings(locals())
        if hostname:
            del self.config["self"]
            del self.config["current_player"]
            self.current_player = int(current_player) or 0
            try:
                if self._debug:
                    print_d(f"Trying {self.config}...")
                self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                self.socket.settimeout(self._TIMEOUT)
                self.socket.connect((socket.gethostbyname(hostname), port))
            except OSError as e:
                print_d(f"Couldn't talk to {self.config} ({e})")
            else:
                result = self.__request(f"login {user} {password}")
                if result != (6 * "*"):
                    raise SqueezeboxError(
                        f"Couldn't log in to squeezebox: response was '{result}'"
                    )
                self.is_connected = True
                self.failures = 0
                print_d(f"Connected to Squeezebox Server! {self}")
                # Reset players (forces reload)
                self.players = []
                self.get_players()

    def get_library_dir(self):
        return self.config["library_dir"]

    def __request(self, line, raw=False, want_reply=True):
        """
        Send a request to the server, if connected, and return its response
        """
        line = line.strip()

        if not (self.is_connected or line.split()[0] == "login"):
            print_d(f"Can't do '{line.split()[0]}' - not connected", self)
            return None

        if self._debug:
            print_(f'>>>> "{line}"')
        try:
            self.socket.send((line + "\n").encode("utf-8"))
            if not want_reply:
                return None
            raw_response = b""
            while not raw_response.endswith(b"\n"):
                raw_response += self.socket.recv(1)
            raw_response = raw_response.decode("utf-8")
        except (TimeoutError, OSError) as e:
            print_w(f"Couldn't communicate with squeezebox ({e})")
            self.failures += 1
            if self.failures >= self._MAX_FAILURES:
                print_w("Too many Squeezebox failures. Disconnecting")
                self.is_connected = False
            return None
        response = (raw_response if raw else unquote(raw_response)).strip()
        if self._debug:
            print_(f'<<<< "{response}"')
        return (
            response[len(line) - 1 :]
            if line.endswith("?")
            else response[len(line) + 1 :]
        )

    def get_players(self):
        """Returns (and caches) a list of the Squeezebox players available"""
        if self.players:
            return self.players
        pairs = self.__request("players 0 99", True).split(" ")

        def demunge(string):
            s = unquote(string)
            cpos = s.index(":")
            return s[0:cpos], s[cpos + 1 :]

        # Do a meaningful URL-unescaping and tuplification for all values
        pairs = [demunge(p) for p in pairs]

        # First element is always count
        count = int(pairs.pop(0)[1])
        self.players = []
        for pair in pairs:
            if pair[0] == "playerindex":
                playerindex = int(pair[1])
                self.players.append(SqueezeboxPlayerSettings())
            else:
                # Don't worry playerindex is always the first entry...
                self.players[playerindex][pair[0]] = pair[1]
        if self._debug:
            print_d("Found %d player(s): %s" % (len(self.players), self.players))
        assert count == len(self.players)
        return self.players

    def player_request(self, line, want_reply=True):
        if not self.is_connected:
            return None
        try:
            return self.__request(
                f"{self.players[self.current_player]['playerid']} {line}",
                want_reply=want_reply,
            )
        except IndexError:
            return None

    def get_version(self):
        if self.is_connected:
            return self.__request("version ?")
        return "(not connected)"

    def play(self):
        """Plays the current song"""
        self.player_request("play")

    def is_stopped(self):
        """Returns whether the player is in any sort of non-playing mode"""
        response = self.player_request("mode ?")
        return "play" != response

    def playlist_play(self, path):
        """Play song immediately"""
        self.player_request(f"playlist play {quote(path)}")

    def playlist_add(self, path):
        self.player_request(f"playlist add {quote(path)}", False)

    def playlist_save(self, name):
        self.player_request(f"playlist save {quote(name)}", False)

    def playlist_clear(self):
        self.player_request("playlist clear", False)

    def playlist_resume(self, name, resume, wipe=False):
        cmd = "playlist resume %s noplay:%d wipePlaylist:%d" % (
            quote(name),
            int(not resume),
            int(wipe),
        )
        self.player_request(cmd, want_reply=False)

    def change_song(self, path):
        """Queue up a song"""
        self.player_request("playlist clear")
        self.player_request(f"playlist insert {quote(path)}")

    def seek_to(self, ms):
        """Seeks the current song to `ms` milliseconds from start"""
        if not self.is_connected:
            return
        if self._debug:
            print_d(
                "Requested %0.2f s, adding drift of %d ms..."
                % (ms / 1000.0, self.delta)
            )
        ms += self.delta
        start = time.time()
        self.player_request("time %d" % round(int(ms) / 1000))
        end = time.time()
        took = (end - start) * 1000
        reported_time = self.get_milliseconds()
        ql_pos = app.player.get_position()
        # Assume 50% of the time taken to complete is response.
        # TODO: Better predictive modelling
        new_delta = ql_pos - reported_time
        self.delta = (self.delta + new_delta) / 2
        if self._debug:
            print_d(
                f"Player at {reported_time / 1000.0:0.0f} "
                f"but QL at {ql_pos / 1000.0:0.2f}."
                f"(Took {took:0.0f} ms). Drift was {new_delta:+0.0f} ms"
            )

    def get_milliseconds(self):
        secs = self.player_request("time ?") or 0
        return float(secs) * 1000.0

    def pause(self):
        self.player_request("pause 1")

    def unpause(self):
        if self.is_stopped():
            self.play()
        ms = app.player.get_position()
        self.seek_to(ms)

    def stop(self):
        self.player_request("stop")

    def __str__(self):
        return str(self.config)



# Copyright 2018 Phoenix Dailey
#           2020 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import random

from gi.repository import Gtk, GLib

from quodlibet import _
from quodlibet import app
from quodlibet.order import OrderInOrder
from quodlibet.order import OrderRemembered
from quodlibet.order.reorder import Reorder
from quodlibet.plugins import PluginConfig
from quodlibet.plugins.playorder import ShufflePlugin
from quodlibet.qltk import Icons
from quodlibet.qltk.notif import Task


pconfig = PluginConfig("shufflebygrouping")
pconfig.defaults.set("grouping", "~grouping~album~albumartist")
pconfig.defaults.set("grouping_filter", "grouping")
pconfig.defaults.set("delay", 0)


class ShuffleByGrouping(ShufflePlugin, OrderRemembered):
    PLUGIN_ID = "Shuffle by Grouping"
    PLUGIN_NAME = _("Shuffle by Grouping")
    PLUGIN_DESC = _(
        "Adds a shuffle mode that groups songs with a common tag "
        "– similar to album shuffles.\n\n"
        "This is useful for shuffling multi-movement classical "
        "pieces, making sure all movements play in order "
        "before shuffling to the next piece."
    )
    PLUGIN_ICON = Icons.MEDIA_PLAYLIST_SHUFFLE
    display_name = _("Shuffle by grouping")
    accelerated_name = _("Shuffle by _grouping")
    priority = Reorder.priority

    def next(self, playlist, current_song):
        return self._next(playlist, current_song)

    def _next(self, playlist, current_song, delay_on=True):
        grouping = str(pconfig.gettext("grouping")).strip()
        grouping_filter = str(pconfig.gettext("grouping_filter")).strip()
        delay = pconfig.getint("delay")

        def same_group(song_iter_a, song_iter_b):
            if song_iter_a is None or song_iter_b is None:
                return False
            song_a = playlist.get_value(song_iter_a)
            song_b = playlist.get_value(song_iter_b)
            if not self._tag_defined(grouping_filter, song_a):
                return False
            if not self._tag_defined(grouping_filter, song_b):
                return False
            return song_a(grouping) == song_b(grouping)

        # Keep track of played songs
        OrderRemembered.next(self, playlist, current_song)
        remaining = OrderRemembered.remaining(self, playlist)

        # Check if playlist is finished or empty
        if not remaining:
            OrderRemembered.reset(self, playlist)
            return None

        # Play next song in current grouping
        next_song = OrderInOrder.next(self, playlist, current_song)
        if same_group(next_song, current_song):
            return next_song

        # Pause for a moment before picking new group
        if delay_on:
            self._resume_after_delay(delay)

        # Pick random song at the start of a new group
        while True:
            song_location = random.choice(list(remaining.keys()))
            new_song = playlist.get_iter(song_location)
            new_song_prev = (
                playlist.get_iter(song_location - 1) if song_location >= 1 else None
            )
            if not same_group(new_song, new_song_prev):
                return new_song

    @staticmethod
    def _resume_after_delay(delay, refresh_rate=20):
        if delay <= 0:
            return
        app.player.paused = True
        delay_timer = GLib.timeout_add(1000 * delay, app.player.play)
        task = Task(
            _("Shuffle by Grouping"),
            _("Waiting to start new group…"),
            stop=lambda: GLib.source_remove(delay_timer),
        )

        def countdown():
            for i in range(int(refresh_rate * delay)):
                task.update(i / (refresh_rate * delay))
                yield True
            task.finish()
            yield False

        GLib.timeout_add(1000 / refresh_rate, next, countdown())

    @staticmethod
    def _tag_defined(tag_name, song):
        if not tag_name:
            return True
        tag_value = song(tag_name)
        return bool(tag_value.strip())

    def next_explicit(self, playlist, current_song):
        return self._next(playlist, current_song, delay_on=False)

    def previous(self, playlist, current_song):
        return OrderRemembered.previous(self, playlist, current_song)

    @classmethod
    def PluginPreferences(cls, window):
        def on_change(widget, config_entry):
            config_value = widget.get_text()
            pconfig.set(config_entry, config_value)

        def on_spin(widget, config_entry):
            config_value = widget.get_value()
            pconfig.set(config_entry, config_value)

        def default_on_click(widget):
            pconfig.reset("grouping")
            pconfig.reset("grouping_filter")
            pconfig.reset("delay")
            grouping_entry.set_text(pconfig.gettext("grouping"))
            grouping_filter_entry.set_text(pconfig.gettext("grouping_filter"))
            delay_spin.set_value(pconfig.getint("delay"))

        def make_label(label_text):
            label = Gtk.Label(label_text, selectable=True)
            label.set_alignment(0.0, 0.5)
            label.set_margin_end(3)
            return label

        vbox = Gtk.VBox(spacing=12)

        grouping_label = make_label(_("Grouping tag:"))
        grouping_entry = Gtk.Entry()
        grouping_entry.connect("changed", on_change, "grouping")
        grouping_entry.set_text(pconfig.gettext("grouping"))
        grouping_entry.set_tooltip_text(_("Tag to group songs by"))

        grouping_filter_label = make_label(_("Filter tag:"))
        grouping_filter_entry = Gtk.Entry()
        grouping_filter_entry.connect("changed", on_change, "grouping_filter")
        grouping_filter_entry.set_text(pconfig.gettext("grouping_filter"))
        grouping_filter_entry.set_tooltip_text(
            _(
                "Grouping is applied only if the filter tag is defined.\n"
                "A song with an undefined filter tag will be treated as\n"
                "a group consisting only of itself. Typically the filter\n"
                "tag should match or partially match the grouping tag."
            )
        )

        delay_label = make_label(_("Delay:"))
        adj = Gtk.Adjustment.new(pconfig.getint("delay"), 0, 3600, 1, 5, 0)
        delay_spin = Gtk.SpinButton(adjustment=adj, climb_rate=0.1, digits=0)
        delay_spin.set_numeric(True)
        delay_spin.connect("value-changed", on_spin, "delay")
        delay_spin.set_tooltip_text(
            _("Time delay in seconds before starting next group")
        )

        table = Gtk.Table(3, 2)
        table.set_col_spacings(6)
        table.set_row_spacings(6)

        table.attach(grouping_label, 0, 1, 0, 1, Gtk.AttachOptions.FILL)
        table.attach(grouping_filter_label, 0, 1, 1, 2, Gtk.AttachOptions.FILL)
        table.attach(delay_label, 0, 1, 2, 3, Gtk.AttachOptions.FILL)

        table.attach(grouping_entry, 1, 2, 0, 1)
        table.attach(grouping_filter_entry, 1, 2, 1, 2)
        table.attach(delay_spin, 1, 2, 2, 3)

        vbox.add(table)

        defaults = Gtk.Button(_("Reset to defaults"))
        defaults.connect("clicked", default_on_click)
        vbox.add(defaults)

        return vbox



# Copyright 2017 Christoph Reiter
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from mutagen.dsf import DSF

from ._id3 import ID3File


class DSFFile(ID3File):
    format = "DSF"

    mimes = ["audio/dsf"]
    Kind = DSF

    def _parse_info(self, info):
        self["~#length"] = info.length
        self["~#bitrate"] = int(info.bitrate / 1000)
        self["~#channels"] = info.channels
        self["~#samplerate"] = info.sample_rate
        self["~#bitdepth"] = info.bits_per_sample


loader = DSFFile
types = [DSFFile]
extensions = [".dsf"]



# Copyright 2004-2005 Joe Wreschnig, Michael Urman, Iñigo Serna
#             2020-23 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import os
import re
import unicodedata
import glob
import shutil

from gi.repository import Gtk, Gdk
from senf import fsn2text, text2fsn

import quodlibet
from quodlibet import qltk
from quodlibet import util
from quodlibet import config
from quodlibet import _

from quodlibet.plugins import PluginManager
from quodlibet.pattern import FileFromPattern
from quodlibet.pattern import ArbitraryExtensionFileFromPattern
from quodlibet.qltk._editutils import FilterPluginBox, FilterCheckButton
from quodlibet.qltk._editutils import EditingPluginHandler
from quodlibet.qltk.views import TreeViewColumn
from quodlibet.qltk.cbes import ComboBoxEntrySave
from quodlibet.qltk.ccb import ConfigCheckButton
from quodlibet.qltk.models import ObjectStore
from quodlibet.qltk import Icons, Button, Frame
from quodlibet.qltk.wlw import WritingWindow
from quodlibet.util import connect_obj
from quodlibet.util.path import strip_win32_incompat_from_path
from quodlibet.util.dprint import print_d, print_e

NBP = os.path.join(quodlibet.get_user_dir(), "lists", "renamepatterns")
NBP_EXAMPLES = """\
<tracknumber>. <title>
<tracknumber|<tracknumber>. ><title>
<tracknumber> - <title>
<tracknumber> - <artist> - <title>
/path/<artist> - <album>/<tracknumber>. <title>
~/<artist>/<album>/<tracknumber> - <title>
<albumartist|<albumartist>|<artist>>/(<~year>) <album>\
/<tracknumber|<tracknumber> - ><title>"""


class SpacesToUnderscores(FilterCheckButton):
    _label = _("Replace spaces with _underscores")
    _section = "rename"
    _key = "spaces"
    _order = 1.0

    def filter(self, original, filename):
        return filename.replace(" ", "_")


class ReplaceColons(FilterCheckButton):
    _label = _("Replace [semi]colon delimiting with hyphens")
    _tooltip = _('e.g. "iv: allegro.flac" → "iv - allegro.flac"')
    _section = "rename"
    _key = "colons"
    _order = 1.05

    def __init__(self):
        super().__init__()
        # If on Windows, force this to be inactive (and hidden)
        if os.name == "nt":
            self.set_active(False)
            self.set_sensitive(False)
            self.set_no_show_all(True)

    def filter(self, original, filename):
        regx = re.compile(r"\s*[:;]\s+\b")
        return regx.sub(" - ", filename)


class StripWindowsIncompat(FilterCheckButton):
    _label = _("Strip _Windows-incompatible characters")
    _section = "rename"
    _key = "windows"
    _order = 1.1

    def __init__(self):
        super().__init__()
        # If on Windows, force this to be inactive (and hidden)
        if os.name == "nt":
            self.set_active(False)
            self.set_sensitive(False)
            self.set_no_show_all(True)

    def filter(self, original, filename):
        return strip_win32_incompat_from_path(filename)


class StripDiacriticals(FilterCheckButton):
    _label = _("Strip _diacritical marks")
    _section = "rename"
    _key = "diacriticals"
    _order = 1.2

    def filter(self, original, filename):
        return "".join(
            filter(
                lambda s: not unicodedata.combining(s),
                unicodedata.normalize("NFKD", filename),
            )
        )


class StripNonASCII(FilterCheckButton):
    _label = _("Strip non-_ASCII characters")
    _section = "rename"
    _key = "ascii"
    _order = 1.3

    def filter(self, original, filename):
        return "".join((s <= "~" and s) or "_" for s in filename)


class Lowercase(FilterCheckButton):
    _label = _("Use only _lowercase characters")
    _section = "rename"
    _key = "lowercase"
    _order = 1.4

    def filter(self, original, filename):
        return filename.lower()


class RenameFilesPluginHandler(EditingPluginHandler):
    from quodlibet.plugins.editing import RenameFilesPlugin

    Kind = RenameFilesPlugin


class Entry:
    def __init__(self, song):
        self.song = song

    new_name = None
    """new name as unicode or None if not set"""

    @property
    def name(self):
        return fsn2text(self.song("~basename"))


RESPONSE_SKIP_ALL = 1


class RenameFiles(Gtk.VBox):
    title = _("Rename Files")
    FILTERS = [
        SpacesToUnderscores,
        ReplaceColons,
        StripWindowsIncompat,
        StripDiacriticals,
        StripNonASCII,
        Lowercase,
    ]
    handler = RenameFilesPluginHandler()
    IMAGE_EXTENSIONS = ["jpg", "jpeg", "png", "bmp"]

    @classmethod
    def init_plugins(cls):
        PluginManager.instance.register_handler(cls.handler)

    def __init__(self, parent, library):
        super().__init__(spacing=6)
        self.__skip_interactive = False
        self.set_border_width(12)

        hbox = Gtk.HBox(spacing=6)
        cbes_defaults = NBP_EXAMPLES.split("\n")
        self.combo = ComboBoxEntrySave(
            NBP,
            cbes_defaults,
            title=_("Path Patterns"),
            edit_title=_("Edit saved patterns…"),
        )
        self.combo.show_all()
        hbox.pack_start(self.combo, True, True, 0)
        self.preview = qltk.Button(_("_Preview"), Icons.VIEW_REFRESH)
        self.preview.show()
        hbox.pack_start(self.preview, False, True, 0)
        self.pack_start(hbox, False, True, 0)
        self.combo.get_child().connect("changed", self._changed)

        model = ObjectStore()
        self.view = Gtk.TreeView(model=model)
        self.view.show()

        sw = Gtk.ScrolledWindow()
        sw.set_shadow_type(Gtk.ShadowType.IN)
        sw.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        sw.add(self.view)
        self.pack_start(sw, True, True, 0)

        self.pack_start(Gtk.VBox(), False, True, 0)

        # rename options
        rename_options = Gtk.HBox()

        # file name options
        filter_box = FilterPluginBox(self.handler, self.FILTERS)
        filter_box.connect("preview", self.__filter_preview)
        filter_box.connect("changed", self.__filter_changed)
        self.filter_box = filter_box

        frame_filename_options = Frame(_("File names"), filter_box)
        frame_filename_options.show_all()
        rename_options.pack_start(frame_filename_options, False, True, 0)

        # album art options
        albumart_box = Gtk.VBox()

        # move art
        moveart_box = Gtk.VBox()
        self.moveart = ConfigCheckButton(
            _("_Move album art"), "rename", "move_art", populate=True
        )
        self.moveart.set_tooltip_text(
            _(
                "See '[albumart] search_filenames' config entry "
                "for which images will be moved"
            )
        )
        self.moveart.show()
        moveart_box.pack_start(self.moveart, False, True, 0)
        self.moveart_overwrite = ConfigCheckButton(
            _("_Overwrite album art at target"),
            "rename",
            "move_art_overwrite",
            populate=True,
        )
        self.moveart_overwrite.show()
        moveart_box.pack_start(self.moveart_overwrite, False, True, 0)
        albumart_box.pack_start(moveart_box, False, True, 0)
        # remove empty
        removeemptydirs_box = Gtk.VBox()
        self.removeemptydirs = ConfigCheckButton(
            _("_Remove empty directories"), "rename", "remove_empty_dirs", populate=True
        )
        self.removeemptydirs.show()
        removeemptydirs_box.pack_start(self.removeemptydirs, False, True, 0)
        albumart_box.pack_start(removeemptydirs_box, False, True, 0)

        frame_albumart_options = Frame(_("Album art"), albumart_box)
        frame_albumart_options.show_all()
        rename_options.pack_start(frame_albumart_options, False, True, 0)

        self.pack_start(rename_options, False, True, 0)

        # Save button
        self.save = Button(_("_Save"), Icons.DOCUMENT_SAVE)
        self.save.show()
        bbox = Gtk.HButtonBox()
        bbox.set_layout(Gtk.ButtonBoxStyle.END)
        bbox.pack_start(self.save, True, True, 0)
        self.pack_start(bbox, False, True, 0)

        render = Gtk.CellRendererText()
        column = TreeViewColumn(title=_("File"))
        column.pack_start(render, True)

        def cell_data_file(column, cell, model, iter_, data):
            entry = model.get_value(iter_)
            cell.set_property("text", entry.name)

        column.set_cell_data_func(render, cell_data_file)

        column.set_sizing(Gtk.TreeViewColumnSizing.AUTOSIZE)
        self.view.append_column(column)

        render = Gtk.CellRendererText()
        render.set_property("editable", True)
        column = TreeViewColumn(title=_("New Name"))
        column.pack_start(render, True)

        def cell_data_new_name(column, cell, model, iter_, data):
            entry = model.get_value(iter_)
            cell.set_property("text", entry.new_name or "")

        column.set_cell_data_func(render, cell_data_new_name)

        column.set_sizing(Gtk.TreeViewColumnSizing.AUTOSIZE)
        self.view.append_column(column)

        connect_obj(self.preview, "clicked", self._preview, None)

        connect_obj(parent, "changed", self.__class__._preview, self)
        connect_obj(self.save, "clicked", self._rename, library)

        render.connect("edited", self.__row_edited)

        for child in self.get_children():
            child.show()

    def __filter_preview(self, *args):
        Gtk.Button.clicked(self.preview)

    def __filter_changed(self, *args):
        self._changed(self.combo.get_child())

    def _changed(self, entry):
        self.save.set_sensitive(False)
        self.preview.set_sensitive(bool(entry.get_text()))

    def __row_edited(self, renderer, path, new):
        path = Gtk.TreePath.new_from_string(path)
        model = self.view.get_model()
        entry = model[path][0]
        if entry.new_name != new:
            entry.new_name = new
            self.preview.set_sensitive(True)
            self.save.set_sensitive(True)
            model.path_changed(path)

    def _rename(self, library):
        model = self.view.get_model()
        win = WritingWindow(self, len(model))
        win.show()
        was_changed = set()
        skip_all = self.__skip_interactive
        self.view.freeze_child_notify()
        should_move_art = config.getboolean("rename", "move_art")
        moveart_sets = {}
        remove_empty_dirs = config.getboolean("rename", "remove_empty_dirs")

        for entry in model.values():
            if entry.new_name is None:
                continue
            song = entry.song
            old_name = entry.name
            old_pathfile = song["~filename"]
            new_name = entry.new_name
            new_pathfile = ""
            # ensure target is a full path
            if os.path.abspath(new_name) != os.path.abspath(
                os.path.join(os.getcwd(), new_name)
            ):
                new_pathfile = new_name
            else:
                # must be a relative pattern, so prefix the path
                new_pathfile = os.path.join(os.path.dirname(old_pathfile), new_name)

            try:
                library.rename(song, text2fsn(new_name), changed=was_changed)
            except Exception:
                util.print_exc()
                if skip_all:
                    continue
                msg = qltk.Message(
                    Gtk.MessageType.ERROR,
                    win,
                    _("Unable to rename file"),
                    _(
                        "Renaming %(old-name)s to %(new-name)s failed. "
                        "Possibly the target file already exists, "
                        "or you do not have permission to make the "
                        "new file or remove the old one."
                    )
                    % {
                        "old-name": util.bold(old_name),
                        "new-name": util.bold(new_name),
                    },
                    buttons=Gtk.ButtonsType.NONE,
                )
                msg.add_button(_("Ignore _All Errors"), RESPONSE_SKIP_ALL)
                msg.add_icon_button(
                    _("_Stop"), Icons.PROCESS_STOP, Gtk.ResponseType.CANCEL
                )
                msg.add_button(_("_Continue"), Gtk.ResponseType.OK)
                msg.set_default_response(Gtk.ResponseType.OK)
                resp = msg.run()
                skip_all |= resp == RESPONSE_SKIP_ALL
                # Preserve old behavior: shift-click is Ignore All
                mods = Gdk.Display.get_default().get_pointer()[3]
                skip_all |= mods & Gdk.ModifierType.SHIFT_MASK
                library.reload(song, changed=was_changed)
                if resp != Gtk.ResponseType.OK and resp != RESPONSE_SKIP_ALL:
                    break

            if should_move_art:
                self._moveart(moveart_sets, old_pathfile, new_pathfile, song)

            if remove_empty_dirs:
                path_old = os.path.dirname(old_pathfile)
                if not os.listdir(path_old):
                    try:
                        os.rmdir(path_old)
                        print_d(f"Removed empty directory: {path_old!r}", self)
                    except Exception:
                        util.print_exc()

            if win.step():
                break

        self.view.thaw_child_notify()
        win.destroy()
        library.changed(was_changed)
        self.save.set_sensitive(False)

    def _moveart(self, art_sets, pathfile_old, pathfile_new, song):
        path_old = os.path.dirname(os.path.realpath(pathfile_old))
        path_new = os.path.dirname(os.path.realpath(pathfile_new))
        if os.path.realpath(path_old) == os.path.realpath(path_new):
            return
        if path_old in art_sets.keys() and not art_sets[path_old]:
            return

        # get art set for path
        images = []
        if path_old in art_sets.keys():
            images = art_sets[path_old]
        else:

            def glob_escape(s):
                for c in "[*?":
                    s = s.replace(c, "[" + c + "]")
                return s

            # generate art set for path
            art_sets[path_old] = images
            path_old_escaped = glob_escape(path_old)
            for suffix in self.IMAGE_EXTENSIONS:
                images.extend(glob.glob(os.path.join(path_old_escaped, "*." + suffix)))
        if images:
            # set not empty yet, (re)process
            filenames = config.getstringlist("albumart", "search_filenames")
            moves = []
            for fn in filenames:
                fn = os.path.join(path_old, fn)
                if "<" in fn:
                    # resolve path
                    fnres = ArbitraryExtensionFileFromPattern(fn).format(song)
                    if fnres in images and fnres not in moves:
                        moves.append(fnres)
                elif "*" in fn:
                    moves.extend(
                        f for f in glob.glob(fn) if f in images and f not in moves
                    )
                elif fn in images and fn not in moves:
                    moves.append(fn)
            if len(moves) > 0:
                overwrite = config.getboolean("rename", "move_art_overwrite")
                for fnmove in moves:
                    try:
                        # existing files safeguarded until move successful,
                        # then deleted if overwrite set
                        fnmoveto = os.path.join(path_new, os.path.split(fnmove)[1])
                        fnmoveto_orig = ""
                        if os.path.exists(fnmoveto):
                            fnmoveto_orig = fnmoveto + ".orig"
                            if not os.path.exists(fnmoveto_orig):
                                os.rename(fnmoveto, fnmoveto_orig)
                            else:
                                suffix = 1
                                while os.path.exists(fnmoveto_orig + "." + str(suffix)):
                                    suffix += 1
                                fnmoveto_orig = fnmoveto_orig + "." + str(suffix)
                                os.rename(fnmoveto, fnmoveto_orig)
                        print_d(f"Renaming image {fnmove!r} to {fnmoveto!r}", self)
                        shutil.move(fnmove, fnmoveto)
                        if overwrite and fnmoveto_orig:
                            os.remove(fnmoveto_orig)
                        images.remove(fnmove)
                    except Exception as e:
                        print_e(f"Couldn't move file ({e})")
                        util.print_exc()

    def _preview(self, songs):
        model = self.view.get_model()
        if songs is None:
            songs = [e.song for e in model.values()]

        pattern_text = self.combo.get_child().get_text()

        try:
            pattern = FileFromPattern(pattern_text)
        except ValueError:
            qltk.ErrorMessage(
                self,
                _("Path is not absolute"),
                _(
                    "The pattern\n\t%s\ncontains / but does not start from root. "
                    "To avoid misnamed folders, "
                    "root your pattern by starting it with / or ~/."
                )
                % (util.bold(pattern_text)),
            ).run()
            return
        else:
            if pattern:
                self.combo.prepend_text(pattern_text)
                self.combo.write(NBP)

        # native paths
        orignames = [song["~filename"] for song in songs]
        newnames = [fsn2text(pattern.format(song)) for song in songs]
        for f in self.filter_box.filters:
            if f.active:
                newnames = f.filter_list(orignames, newnames)

        model.clear()
        for song, newname in zip(songs, newnames, strict=False):
            entry = Entry(song)
            entry.new_name = newname
            model.append(row=[entry])

        self.preview.set_sensitive(False)
        self.save.set_sensitive(bool(pattern_text))
        for song in songs:
            if not song.is_file:
                self.set_sensitive(False)
                break
        else:
            self.set_sensitive(True)

    @property
    def test_mode(self):
        return self.__skip_interactive

    @test_mode.setter
    def test_mode(self, value):
        self.__skip_interactive = value



# Copyright 2021 Joschua Gandert
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
from dataclasses import dataclass

from gi.repository import Gtk, Pango
from typing import Generic, TypeVar
from collections.abc import Callable

from quodlibet.qltk.models import ObjectStore

from quodlibet.qltk.views import HintedTreeView

from quodlibet.qltk import Icons

from quodlibet import _, app

from quodlibet.qltk.window import PersistentWindowMixin, Dialog


MATCH_DESC = _(
    "Check if the columns on the left side approximately match the ones on "
    "the right side. If they don't, you can change the order here (use _ "
    "for rows that shouldn't be matched):"
)

T = TypeVar("T")


@dataclass
class ColumnSpec(Generic[T]):
    """Simple data class for specifying the behavior and content of a column"""

    title: str
    cell_text_getter: Callable[[T], str]
    is_resizable: bool = True


# We're using a Dialog, since a ConfirmationPrompt looks really ugly at such a width
class MatchListsDialog(Dialog, PersistentWindowMixin, Generic[T]):
    """
    A prompt whose run method returns the chosen order, or an empty list if the user
    pressed cancel.
    """

    def __init__(
        self,
        a_items: list[T],
        b_items: list[T],
        b_order: list[int | None],
        columns: list[ColumnSpec[T]],
        title: str,
        ok_button_text: str,
        ok_button_icon: str = Icons.DOCUMENT_SAVE,
        description: str = MATCH_DESC,
        parent=app.window,
        id_for_window_tracking: str | None = None,
    ):
        super().__init__(
            title=title, transient_for=parent, modal=True, destroy_with_parent=True
        )

        # A lot of information to display, so make resizable and maximize
        self.set_resizable(True)
        self.set_default_size(750, 500)
        self.maximize()

        if id_for_window_tracking is not None:
            self.enable_window_tracking(id_for_window_tracking)

        vb = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=8)
        self.get_content_area().pack_start(vb, True, True, 0)
        vb.set_spacing(24)
        self.set_border_width(5)

        desc_lbl = Gtk.Label(f"\n{description}\n", wrap=True)
        vb.pack_start(desc_lbl, False, False, 0)

        self.add_button(_("_Cancel"), Gtk.ResponseType.REJECT)
        self.add_icon_button(ok_button_text, ok_button_icon, Gtk.ResponseType.OK)

        order_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=8)

        lbl = Gtk.Label(_("Right side order:"))
        order_box.pack_start(lbl, False, True, 0)

        self.order_entry = Gtk.Entry()
        order_box.pack_start(self.order_entry, True, True, 0)

        vb.pack_start(order_box, False, True, 1)

        sw = Gtk.ScrolledWindow()
        sw.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)

        vb.pack_start(sw, True, True, 1)

        tree = MatchListsTreeView(a_items, b_items, columns)
        self._tree = tree
        sw.add(tree)
        tree.b_order = b_order

        default_order_text = ", ".join(
            ("_" if i is None else str(i + 1)) for i in b_order
        )
        self.order_entry.set_text(default_order_text)
        legal_characters = set(default_order_text)
        legal_characters.add("_")

        def changed_order_entry(widget):
            text = widget.get_text()

            # remove illegal characters
            new_text = "".join(c for c in text if c in legal_characters)
            widget.set_text(new_text)

            num_tracks = len(b_items)
            reordered = one_indexed_csv_to_unique_indices(new_text, num_tracks)
            if reordered:
                tree.b_order = reordered

        self.order_entry.connect("changed", changed_order_entry)

    def run(self, destroy=True) -> list[int | None]:
        self.show_all()
        resp = super().run()
        if destroy:
            self.destroy()
        return self.order if resp == Gtk.ResponseType.OK else []

    @property
    def order(self):
        return self._tree.b_order


def one_indexed_csv_to_unique_indices(
    text,
    target_length,
    char_for_none_matching="_",
    require_target_length_elements=False,
):
    """
    :return: List of indices from the comma separated list of one-indexed numbers.
             List will be empty if any index is repeated or out of bounds.

    By default '_' can be used to specify that a position has no match, which
    will be represented in the returned list with None.
    """

    # remove trailing commas, then split
    nums = text.strip(", ").split(",")
    if require_target_length_elements and len(nums) != target_length:
        return []

    reordered = []
    for n in nums:
        n = n.strip()

        if n == char_for_none_matching:
            i = None
        else:
            try:
                i = int(n) - 1
            except ValueError:
                return []

            # i is not in range and/or not unique
            if i < 0 or i >= target_length or i in reordered:
                return []

        reordered.append(i)

    return reordered


class MatchListsTreeView(HintedTreeView, Generic[T]):
    _b_order: list[int | None]

    def __init__(
        self, a_items: list[T], b_items: list[T], columns: list[ColumnSpec[T]]
    ):
        self.model = ObjectStore()
        self.model.append_many(a_items)
        self._b_items = b_items

        super().__init__(self.model)
        self.set_headers_clickable(False)
        self.set_rules_hint(True)
        self.set_reorderable(False)
        self.get_selection().set_mode(Gtk.SelectionMode.NONE)

        def show_id(col, cell, model, itr, data):
            idx = model.get_path(itr)[0]
            imp_idx = self._b_order[idx]
            num = "_" if imp_idx is None else imp_idx + 1
            cell.set_property("markup", f'<span weight="bold">{num}</span>')

        def df_for_a_items(a_attr_getter):
            def data_func(col, cell, model, itr, data):
                a_item = model[itr][0]
                text = ""
                if a_item is not None:
                    text = a_attr_getter(a_item)
                cell.set_property("text", text)

            return data_func

        def df_for_b_items(b_attr_getter):
            def data_func(col, cell, model, itr, data):
                self._set_text(model, itr, cell, b_attr_getter)

            return data_func

        for c in columns:
            self._add_col(c.title, df_for_a_items(c.cell_text_getter), c.is_resizable)

        self._add_col("#", show_id, False)

        for c in columns:
            self._add_col(c.title, df_for_b_items(c.cell_text_getter), c.is_resizable)

        self._b_order = []  # Initialize the backing field of b_order
        self.b_order = list(range(len(b_items)))  # Update it and rows

        self.update_b_items(b_items)

    def _add_col(self, title, func, resize):
        render = Gtk.CellRendererText()
        render.set_property("ellipsize", Pango.EllipsizeMode.END)
        col = Gtk.TreeViewColumn(title, render)
        col.set_cell_data_func(render, func)
        col.set_resizable(resize)
        col.set_expand(resize)
        self.append_column(col)

    def _set_text(self, model, itr, cell, get_attr):
        idx = model.get_path(itr)[0]
        text = ""
        if idx < len(self._b_order):
            it_idx = self._b_order[idx]
            if it_idx is not None:
                text = get_attr(self._b_items[it_idx])
        cell.set_property("text", text)

    def update_b_items(self, b_items: list[T]):
        """
        Updates the TreeView, handling results with a different number of b_items than
        there are a_items.
        """
        self._b_items = b_items

        for _i in range(len(self.model), len(b_items)):
            self.model.append((None,))

        for _i in range(len(self.model), len(b_items), -1):
            if self.model[-1][0] is not None:
                break
            itr = self.model.get_iter_from_string(str(len(self.model) - 1))
            self.model.remove(itr)

        self._rows_changed()
        self.columns_autosize()

    def _rows_changed(self):
        for row in self.model:
            self.model.row_changed(row.path, row.iter)

    @property
    def b_order(self) -> list[int | None]:
        return list(self._b_order)

    @b_order.setter
    def b_order(self, order: list[int | None]):
        """
        Supports a partial order list. For example, if there are 5 elements in the
        b_items list, you could supply [4, 1, 2]. This will result in an ascending order
        for the last 2 rows, so [0, 3].
        """
        if order == self._b_order:
            return

        b_len = len(self._b_items)
        if len(order) < b_len:
            # add missing indices
            for i in range(b_len):
                if i not in order:
                    order.append(i)

        while len(order) < len(self.model):
            order.append(None)

        self._b_order = order
        self._rows_changed()



# Copyright 2004-2005 Joe Wreschnig, Michael Urman, Iñigo Serna
#           2012 Christoph Reiter
#           2012-2023 Nick Boultbee
#           2017 Uriel Zajaczkovski
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import os

from gi.repository import Gtk, Gdk, GLib, Gio, GObject
from senf import uri2fsn, fsnative, path2fsn

import quodlibet

from quodlibet import browsers
from quodlibet import config
from quodlibet import const
from quodlibet import formats
from quodlibet import qltk
from quodlibet import util
from quodlibet import app
from quodlibet import ngettext
from quodlibet import _
from quodlibet.qltk.paned import ConfigRHPaned

from quodlibet.qltk.appwindow import AppWindow
from quodlibet.update import UpdateDialog
from quodlibet.formats.remote import RemoteFile
from quodlibet.qltk.browser import LibraryBrowser, FilterMenu
from quodlibet.qltk.chooser import choose_folders, choose_files, create_chooser_filter
from quodlibet.qltk.controls import PlayControls
from quodlibet.qltk.cover import CoverImage
from quodlibet.qltk.getstring import GetStringDialog
from quodlibet.qltk.bookmarks import EditBookmarks
from quodlibet.qltk.shortcuts import show_shortcuts
from quodlibet.qltk.info import SongInfo
from quodlibet.qltk.information import Information
from quodlibet.qltk.msg import ErrorMessage, WarningMessage
from quodlibet.qltk.notif import StatusBar, TaskController
from quodlibet.qltk.playorder import (
    PlayOrderWidget,
    RepeatSongForever,
    RepeatListForever,
)
from quodlibet.qltk.pluginwin import PluginWindow
from quodlibet.qltk.properties import SongProperties
from quodlibet.qltk.prefs import PreferencesWindow
from quodlibet.qltk.queue import QueueExpander
from quodlibet.qltk.songlist import SongList, get_columns, set_columns
from quodlibet.qltk.songmodel import PlaylistMux
from quodlibet.qltk.x import RVPaned, Align, ScrolledWindow, Action
from quodlibet.qltk.x import ToggleAction, RadioAction, HighlightToggleButton
from quodlibet.qltk.x import SeparatorMenuItem, MenuItem
from quodlibet.qltk import Icons
from quodlibet.qltk.about import AboutDialog
from quodlibet.util import copool, connect_destroy, connect_after_destroy
from quodlibet.util.library import get_scan_dirs
from quodlibet.util import connect_obj, print_d
from quodlibet.util.library import background_filter, scan_library
from quodlibet.util.path import uri_is_valid
from quodlibet.qltk.window import PersistentWindowMixin, Window, on_first_map
from quodlibet.qltk.songlistcolumns import CurrentColumn


class PlayerOptions(GObject.Object):
    """Provides a simplified interface for playback options.

    This currently provides a limited view on the play order state which is
    useful for external interfaces (mpd, mpris, etc.) and for reducing
    the dependency on the state holding widgets in the main window.

    Usable as long as the main window is not destroyed, or until `destroy()`
    is called.
    """

    __gproperties__ = {
        "shuffle": (
            bool,
            "",
            "",
            False,
            GObject.ParamFlags.READABLE | GObject.ParamFlags.WRITABLE,
        ),
        "repeat": (
            bool,
            "",
            "",
            False,
            GObject.ParamFlags.READABLE | GObject.ParamFlags.WRITABLE,
        ),
        "single": (
            bool,
            "",
            "",
            False,
            GObject.ParamFlags.READABLE | GObject.ParamFlags.WRITABLE,
        ),
        "stop-after": (
            bool,
            "",
            "",
            False,
            GObject.ParamFlags.READABLE | GObject.ParamFlags.WRITABLE,
        ),
    }

    def __init__(self, window):
        """`window` is a QuodLibetWindow"""

        super().__init__()

        self._stop_after = window.stop_after
        self._said = self._stop_after.connect(
            "toggled", lambda *x: self.notify("stop-after")
        )

        def order_changed(*args):
            self.notify("shuffle")
            self.notify("single")

        self._order_widget = window.order
        self._oid = self._order_widget.connect("changed", order_changed)

        window.connect("destroy", self._window_destroy)

    def _window_destroy(self, window):
        self.destroy()

    def destroy(self):
        if self._order_widget:
            self._order_widget.disconnect(self._oid)
            self._order_widget = None
        if self._stop_after:
            self._stop_after.disconnect(self._said)
            self._stop_after = None

    def do_get_property(self, param):
        return getattr(self, param.name.replace("-", "_"))

    def do_set_property(self, param, value):
        setattr(self, param.name.replace("-", "_"), value)

    @property
    def single(self):
        """If only the current song is considered as next track

        When `repeat` is False the playlist will end after this song finishes.
        When `repeat` is True the current song will be replayed.
        """

        return (
            self._order_widget
            and self._order_widget.repeated
            and self._order_widget.repeater is RepeatSongForever
        )

    @single.setter
    def single(self, value):
        if value:
            self.repeat = True
            self._order_widget.repeater = RepeatSongForever
        else:
            self.repeat = False
            self._order_widget.repeater = RepeatListForever

    @property
    def shuffle(self):
        """If a shuffle-like (reordering) play order is active"""

        return self._order_widget.shuffled

    @shuffle.setter
    def shuffle(self, value):
        self._order_widget.shuffled = value

    @property
    def repeat(self):
        """If the player is in some kind of repeat mode"""

        return self._order_widget.repeated

    @repeat.setter
    def repeat(self, value):
        print_d(f"setting repeated to {value}")
        self._order_widget.repeated = value

    @property
    def stop_after(self):
        """If the player will pause after the current song ends"""

        return self._stop_after.get_active()

    @stop_after.setter
    def stop_after(self, value):
        self._stop_after.set_active(value)


class DockMenu(Gtk.Menu):
    """Menu used for the OSX dock and the tray icon"""

    def __init__(self, app):
        super().__init__()

        player = app.player

        play_item = MenuItem(_("_Play"), Icons.MEDIA_PLAYBACK_START)
        play_item.connect("activate", self._on_play, player)
        pause_item = MenuItem(_("P_ause"), Icons.MEDIA_PLAYBACK_PAUSE)
        pause_item.connect("activate", self._on_pause, player)
        self.append(play_item)
        self.append(pause_item)

        previous = MenuItem(_("Pre_vious"), Icons.MEDIA_SKIP_BACKWARD)
        previous.connect("activate", lambda *args: player.previous())
        self.append(previous)

        next_ = MenuItem(_("_Next"), Icons.MEDIA_SKIP_FORWARD)
        next_.connect("activate", lambda *args: player.next())
        self.append(next_)

        browse = qltk.MenuItem(_("_Browse Library"), Icons.EDIT_FIND)
        browse_sub = Gtk.Menu()
        for Kind in browsers.browsers:
            i = Gtk.MenuItem(label=Kind.accelerated_name, use_underline=True)
            connect_obj(
                i, "activate", LibraryBrowser.open, Kind, app.library, app.player
            )
            browse_sub.append(i)

        browse.set_submenu(browse_sub)
        self.append(SeparatorMenuItem())
        self.append(browse)

        self.show_all()
        self.hide()

    def _on_play(self, item, player):
        player.paused = False

    def _on_pause(self, item, player):
        player.paused = True


class MainSongList(SongList):
    """SongList for the main browser's displayed songs."""

    _activated = False

    def __init__(self, library, player):
        super().__init__(library, player, update=True)
        self.set_first_column_type(CurrentColumn)

        self.connect("row-activated", self.__select_song, player)

        # ugly.. so the main window knows if the next song-started
        # comes from an row-activated or anything else.
        def reset_activated(*args):
            self._activated = False

        connect_after_destroy(player, "song-started", reset_activated)

        self.connect("orders-changed", self.__orders_changed)

    def __orders_changed(self, *args):
        l = []
        for tag, reverse in self.get_sort_orders():
            l.append("%d%s" % (int(reverse), tag))
        config.setstringlist("memory", "sortby", l)

    def __select_song(self, widget, indices, col, player):
        self._activated = True
        iter = self.model.get_iter(indices)
        if player.go_to(iter, explicit=True, source=self.model):
            player.paused = False


class TopBar(Gtk.Toolbar):
    def __init__(self, parent, player, library):
        super().__init__()

        # play controls
        control_item = Gtk.ToolItem()
        self.insert(control_item, 0)
        t = PlayControls(player, library.librarian)
        self.volume = t.volume

        # only restore the volume in case it is managed locally, otherwise
        # this could affect the system volume
        if not player.has_external_volume:
            player.volume = config.getfloat("memory", "volume")

        connect_destroy(player, "notify::volume", self._on_volume_changed)
        control_item.add(t)

        self.insert(Gtk.SeparatorToolItem(), 1)

        info_item = Gtk.ToolItem()
        self.insert(info_item, 2)
        info_item.set_expand(True)

        box = Gtk.Box(spacing=6)
        info_item.add(box)
        qltk.add_css(self, "GtkToolbar {padding: 3px;}")

        self._pattern_box = Gtk.VBox(spacing=3)

        # song text
        info_pattern_path = os.path.join(quodlibet.get_user_dir(), "songinfo")
        text = SongInfo(library.librarian, player, info_pattern_path)
        self._pattern_box.pack_start(text, True, True, 0)
        box.pack_start(self._pattern_box, True, True, 0)

        # cover image
        self.image = CoverImage(resize=True)
        connect_destroy(player, "song-started", self.__new_song)

        # FIXME: makes testing easier
        if app.cover_manager:
            connect_destroy(
                app.cover_manager, "cover-changed", self.__song_art_changed, library
            )

        box.pack_start(Align(self.image, top=3, right=3), False, True, 0)

        # On older Gtk+ (3.4, at least)
        # setting a margin on CoverImage leads to errors and result in the
        # QL window not being visible for some reason.
        assert self.image.props.margin == 0

        for child in self.get_children():
            child.show_all()

        context = self.get_style_context()
        context.add_class("primary-toolbar")

    def set_seekbar_widget(self, widget):
        children = self._pattern_box.get_children()
        if len(children) > 1:
            self._pattern_box.remove(children[-1])

        if widget:
            self._pattern_box.pack_start(widget, False, True, 0)

    def _on_volume_changed(self, player, *args):
        config.set("memory", "volume", str(player.volume))

    def __new_song(self, player, song):
        self.image.set_song(song)

    def __song_art_changed(self, player, songs, library):
        self.image.refresh()


class QueueButton(HighlightToggleButton):
    def __init__(self):
        # XXX: view-list isn't part of the fdo spec, so fall back t justify..
        gicon = Gio.ThemedIcon.new_from_names(
            [
                "view-list-symbolic",
                "format-justify-fill-symbolic",
                "view-list",
                "format-justify",
            ]
        )
        image = Gtk.Image.new_from_gicon(gicon, Gtk.IconSize.SMALL_TOOLBAR)

        super().__init__(image=image)

        self.set_name("ql-queue-button")
        qltk.add_css(
            self,
            """
            #ql-queue-button {
                padding: 0px;
            }
        """,
        )
        self.set_size_request(26, 26)

        self.set_tooltip_text(_("Toggle queue visibility"))


class StatusBarBox(Gtk.HBox):
    def __init__(self, play_order, queue):
        super().__init__(spacing=6)
        self.pack_start(play_order, False, True, 0)
        self.statusbar = StatusBar(TaskController.default_instance)
        self.pack_start(self.statusbar, True, True, 0)
        queue_button = QueueButton()
        queue_button.bind_property(
            "active", queue, "visible", GObject.BindingFlags.BIDIRECTIONAL
        )
        queue_button.props.active = queue.props.visible

        self.pack_start(queue_button, False, True, 0)


class PlaybackErrorDialog(ErrorMessage):
    def __init__(self, parent, player_error):
        def add_full_stop(s):
            return s and s.rstrip(".") + "."

        description = add_full_stop(util.escape(player_error.short_desc))
        details = add_full_stop(util.escape(player_error.long_desc or ""))
        if details:
            description += " " + details

        super().__init__(parent, _("Playback Error"), description)


class ConfirmLibDirSetup(WarningMessage):
    RESPONSE_SETUP = 1

    def __init__(self, parent):
        title = _("Set up library directories?")
        description = _(
            "You don't have any music library set up. Would you like to do that now?"
        )

        super().__init__(parent, title, description, buttons=Gtk.ButtonsType.NONE)

        self.add_button(_("_Not Now"), Gtk.ResponseType.CANCEL)
        self.add_button(_("_Set Up"), self.RESPONSE_SETUP)
        self.set_default_response(Gtk.ResponseType.CANCEL)


MENU = """
<ui>
  <menubar name='Menu'>

    <menu action='File'>
      <menuitem action='AddFolders' always-show-image='true'/>
      <menuitem action='AddFiles' always-show-image='true'/>
      <menuitem action='AddLocation' always-show-image='true'/>
      <separator/>
      <menuitem action='Preferences' always-show-image='true'/>
      <menuitem action='Plugins' always-show-image='true'/>
      <separator/>
      <menuitem action='RefreshLibrary' always-show-image='true'/>
      <separator/>
      <menuitem action='Quit' always-show-image='true'/>
    </menu>

    <menu action='Song'>
      <menuitem action='EditBookmarks' always-show-image='true'/>
      <menuitem action='EditTags' always-show-image='true'/>
      <separator/>
      <menuitem action='Information' always-show-image='true'/>
      <separator/>
      <menuitem action='Jump' always-show-image='true'/>
    </menu>

    <menu action='Control'>
      <menuitem action='Previous' always-show-image='true'/>
      <menuitem action='PlayPause' always-show-image='true'/>
      <menuitem action='Next' always-show-image='true'/>
      <menuitem action='Stop' always-show-image='true'/>
      <menuitem action='StopAfter' always-show-image='true'/>
    </menu>

    <menu action='Browse'>
      %(filters_menu)s
      <separator/>
      <menu action='BrowseLibrary' always-show-image='true'>
        %(browsers)s
      </menu>
      <separator />

      %(views)s
    </menu>

    <menu action='Help'>
      <menuitem action='OnlineHelp' always-show-image='true'/>
      <menuitem action='Shortcuts' always-show-image='true'/>
      <menuitem action='SearchHelp' always-show-image='true'/>
      <separator/>
      <menuitem action='CheckUpdates' always-show-image='true'/>
      <menuitem action='About' always-show-image='true'/>
    </menu>

  </menubar>
</ui>
"""


def secondary_browser_menu_items():
    items = (
        _browser_items("Browser") + ["<separator />"] + _browser_items("Browser", True)
    )
    return "\n".join(items)


def browser_menu_items():
    items = _browser_items("View") + ["<separator />"] + _browser_items("View", True)
    return "\n".join(items)


def _browser_items(prefix, external=False):
    return [
        f"<menuitem action='{prefix}{kind.__name__}'/>"
        for kind in browsers.browsers
        if kind.uses_main_library ^ external
    ]


(DND_URI_LIST,) = range(1)


class SongListPaned(RVPaned):
    def __init__(self, song_scroller, qexpander):
        super().__init__()

        self.pack1(song_scroller, resize=True, shrink=False)
        self.pack2(qexpander, resize=True, shrink=False)

        self.set_relative(config.getfloat("memory", "queue_position", 0.75))
        self.connect("notify::position", self._changed, "memory", "queue_position")

        self._handle_position = self.get_relative()
        qexpander.connect("notify::visible", self._expand_or)
        qexpander.connect("notify::expanded", self._expand_or)
        qexpander.connect("draw", self._check_minimize)

        self.connect("button-press-event", self._on_button_press)
        self.connect("notify", self._moved_pane_handle)

    @property
    def _expander(self):
        return self.get_child2()

    def _on_button_press(self, pane, event):
        # If we start to drag the pane handle while the
        # queue expander is unexpanded, expand it and move the handle
        # to the bottom, so we can 'drag' the queue out

        if event.window != pane.get_handle_window():
            return False

        if not self._expander.get_expanded():
            self._expander.set_expanded(True)
            pane.set_relative(1.0)
        return False

    def _expand_or(self, widget, prop):
        if self._expander.get_property("expanded"):
            self.set_relative(self._handle_position)

    def _moved_pane_handle(self, widget, prop):
        if self._expander.get_property("expanded"):
            self._handle_position = self.get_relative()

    def _check_minimize(self, *args):
        if not self._expander.get_property("expanded"):
            p_max = self.get_property("max-position")
            p_cur = self.get_property("position")
            if p_max != p_cur:
                self.set_property("position", p_max)

    def _changed(self, widget, event, section, option):
        if self._expander.get_expanded() and self.get_property("position-set"):
            config.set(section, option, str(self.get_relative()))


class QuodLibetWindow(Window, PersistentWindowMixin, AppWindow):
    def __init__(self, library, player, headless=False, restore_cb=None):
        super().__init__(dialog=False)

        self.__destroyed = False
        self.__update_title(player)
        self.set_default_size(600, 480)

        main_box = Gtk.VBox()
        self.add(main_box)
        self.side_book = qltk.Notebook()

        # get the playlist up before other stuff
        self.songlist = MainSongList(library, player)
        self.songlist.connect("key-press-event", self.__songlist_key_press)
        self.songlist.connect_after(
            "drag-data-received", self.__songlist_drag_data_recv
        )
        self.song_scroller = ScrolledWindow()
        self.song_scroller.set_policy(
            Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC
        )
        self.song_scroller.set_shadow_type(Gtk.ShadowType.IN)
        self.song_scroller.add(self.songlist)

        self.qexpander = QueueExpander(library, player)
        self.qexpander.set_no_show_all(True)
        self.qexpander.set_visible(config.getboolean("memory", "queue"))

        def on_queue_visible(qex, param):
            config.set("memory", "queue", str(qex.get_visible()))

        self.qexpander.connect("notify::visible", on_queue_visible)

        self.playlist = PlaylistMux(player, self.qexpander.model, self.songlist.model)

        self.__player = player
        # create main menubar, load/restore accelerator groups
        self.__library = library
        ui = self.__create_menu(player, library)
        accel_group = ui.get_accel_group()
        self.add_accel_group(accel_group)

        def scroll_and_jump(*args):
            self.__jump_to_current(True, None, True)

        keyval, mod = Gtk.accelerator_parse("<Primary><shift>J")
        accel_group.connect(keyval, mod, 0, scroll_and_jump)

        # custom accel map
        accel_fn = os.path.join(quodlibet.get_user_dir(), "accels")
        Gtk.AccelMap.load(accel_fn)
        # save right away so we fill the file with example comments of all
        # accels
        Gtk.AccelMap.save(accel_fn)

        menubar = ui.get_widget("/Menu")

        # Since https://git.gnome.org/browse/gtk+/commit/?id=b44df22895c79
        # toplevel menu items show an empty 16x16 image. While we don't
        # need image items there UIManager creates them by default.
        # Work around by removing the empty GtkImages
        for child in menubar.get_children():
            if isinstance(child, Gtk.ImageMenuItem):
                child.set_image(None)

        main_box.pack_start(menubar, False, True, 0)

        top_bar = TopBar(self, player, library)
        main_box.pack_start(top_bar, False, True, 0)
        self.top_bar = top_bar

        self.__browserbox = Align(top=3, bottom=3)
        self.__paned = paned = ConfigRHPaned("memory", "sidebar_pos", 0.25)
        paned.pack1(self.__browserbox, resize=True)
        # We'll pack2 when necessary (when the first sidebar plugin is set up)

        main_box.pack_start(paned, True, True, 0)

        play_order = PlayOrderWidget(self.songlist.model, player)
        statusbox = StatusBarBox(play_order, self.qexpander)
        self.order = play_order
        self.statusbar = statusbox.statusbar

        align = Align(statusbox, top=1, bottom=4, left=6, right=6)
        main_box.pack_start(align, False, True, 0)

        self.songpane = SongListPaned(self.song_scroller, self.qexpander)
        self.songpane.show_all()

        try:
            orders = []
            for e in config.getstringlist("memory", "sortby", []):
                orders.append((e[1:], int(e[0])))
        except ValueError:
            pass
        else:
            self.songlist.set_sort_orders(orders)

        self.browser = None
        self.ui = ui

        main_box.show_all()

        self._playback_error_dialog = None
        connect_destroy(player, "song-started", self.__song_started)
        connect_destroy(player, "paused", self.__update_paused, True)
        connect_destroy(player, "unpaused", self.__update_paused, False)
        # make sure we redraw all error indicators before opening
        # a dialog (blocking the main loop), so connect after default handlers
        connect_after_destroy(player, "error", self.__player_error)
        # connect after to let SongTracker update stats
        connect_after_destroy(player, "song-ended", self.__song_ended)

        # set at least the playlist. the song should be restored
        # after the browser emits the song list
        player.setup(self.playlist, None, 0)
        self.__restore_cb = restore_cb
        self.__first_browser_set = True

        restore_browser = not headless
        try:
            self._select_browser(
                self, config.get("memory", "browser"), library, player, restore_browser
            )
        except Exception:
            config.set("memory", "browser", browsers.name(browsers.default))
            config.save()
            raise

        self.songlist.connect("popup-menu", self.__songs_popup_menu)
        self.songlist.connect("columns-changed", self.__cols_changed)
        self.songlist.connect("columns-changed", self.__hide_headers)
        self.songlist.info.connect("changed", self.__set_totals)

        lib = library.librarian
        connect_destroy(lib, "changed", self.__song_changed, player)

        targets = [("text/uri-list", Gtk.TargetFlags.OTHER_APP, DND_URI_LIST)]
        targets = [Gtk.TargetEntry.new(*t) for t in targets]

        self.drag_dest_set(Gtk.DestDefaults.ALL, targets, Gdk.DragAction.COPY)
        self.connect("drag-data-received", self.__drag_data_received)

        if not headless:
            on_first_map(self, self.__configure_scan_dirs, library)

        if config.getboolean("library", "refresh_on_start"):
            self.__rebuild(None, False)

        self.connect("key-press-event", self.__key_pressed, player)

        self.connect("destroy", self.__destroy)

        self.enable_window_tracking("quodlibet")

    def hide_side_book(self):
        self.side_book.hide()

    def add_sidebar(self, box, name):
        vbox = Gtk.Box(margin=0)
        vbox.pack_start(box, True, True, 0)
        vbox.show()
        if self.side_book_empty:
            self.add_sidebar_to_layout(self.side_book)
        self.side_book.append_page(vbox, label=name)
        self.side_book.set_tab_detachable(vbox, False)
        self.side_book.show_all()
        return vbox

    def remove_sidebar(self, widget):
        self.side_book.remove_page(self.side_book.page_num(widget))
        if self.side_book_empty:
            print_d("Hiding sidebar")
            self.__paned.remove(self.__paned.get_children()[1])

    def add_sidebar_to_layout(self, widget):
        print_d("Recreating sidebar")
        align = Align(widget, top=6, bottom=3)
        self.__paned.pack2(align, shrink=True)
        align.show_all()

    @property
    def side_book_empty(self):
        return not self.side_book.get_children()

    def set_seekbar_widget(self, widget):
        """Add an alternative seek bar widget.

        Args:
            widget (Gtk.Widget): a new widget or None to remove the current one
        """

        self.top_bar.set_seekbar_widget(widget)

    def set_as_osx_window(self, osx_app):
        assert osx_app

        self._dock_menu = DockMenu(app)
        osx_app.set_dock_menu(self._dock_menu)

        menu = self.ui.get_widget("/Menu")
        menu.hide()
        osx_app.set_menu_bar(menu)
        # Reparent some items to the "Application" menu
        item = self.ui.get_widget("/Menu/Help/About")
        osx_app.insert_app_menu_item(item, 0)
        osx_app.insert_app_menu_item(Gtk.SeparatorMenuItem(), 1)
        item = self.ui.get_widget("/Menu/File/Preferences")
        osx_app.insert_app_menu_item(item, 2)
        quit_item = self.ui.get_widget("/Menu/File/Quit")
        quit_item.hide()

    def get_is_persistent(self):
        return True

    def open_file(self, filename):
        assert isinstance(filename, fsnative)

        song = self.__library.add_filename(filename, add=False)
        if song is not None:
            if self.__player.go_to(song):
                self.__player.paused = False
            return True
        return False

    def enqueue(self, songs, limit=0):
        """Append `songs` to the queue

        Ask for confimation if the number of songs exceeds `limit`.
        """

        if len(songs) > limit:
            dialog = ConfirmEnqueue(self, len(songs))
            if dialog.run() != Gtk.ResponseType.YES:
                return

        self.playlist.enqueue(songs)

    def __player_error(self, player, song, player_error):
        # it's modal, but mmkeys etc. can still trigger new ones
        if self._playback_error_dialog:
            self._playback_error_dialog.destroy()
        dialog = PlaybackErrorDialog(self, player_error)
        self._playback_error_dialog = dialog
        dialog.run()
        self._playback_error_dialog = None

    def __configure_scan_dirs(self, library):
        """Get user to configure scan dirs, if none is set up"""
        if (
            not get_scan_dirs()
            and not len(library)
            and quodlibet.is_first_session("quodlibet")
        ):
            print_d("Couldn't find any scan dirs")

            resp = ConfirmLibDirSetup(self).run()
            if resp == ConfirmLibDirSetup.RESPONSE_SETUP:
                prefs = PreferencesWindow(self)
                prefs.set_page("library")
                prefs.show()

    def __keyboard_shortcuts(self, action):
        show_shortcuts(self)

    def __edit_bookmarks(self, librarian, player):
        if player.song:
            window = EditBookmarks(self, librarian, player)
            window.show()

    def __key_pressed(self, widget, event, player):
        if not player.song:
            return None

        def seek_relative(seconds):
            current = player.get_position()
            current += seconds * 1000
            current = min(player.song("~#length") * 1000 - 1, current)
            current = max(0, current)
            player.seek(current)

        if qltk.is_accel(event, "<alt>Right"):
            seek_relative(10)
            return True
        if qltk.is_accel(event, "<alt>Left"):
            seek_relative(-10)
            return True
        return None

    def __destroy(self, *args):
        self.playlist.destroy()

        # The tray icon plugin tries to unhide QL because it gets disabled
        # on Ql exit. The window should stay hidden after destroy.
        self.show = lambda: None
        self.present = self.show

    def __drag_data_received(self, widget, ctx, x, y, sel, tid, etime):
        assert tid == DND_URI_LIST

        uris = sel.get_uris()

        dirs = []
        error = False
        for uri in uris:
            try:
                filename = uri2fsn(uri)
            except ValueError:
                filename = None

            if filename is not None:
                loc = os.path.normpath(filename)
                if os.path.isdir(loc):
                    dirs.append(loc)
                else:
                    loc = os.path.realpath(loc)
                    if loc not in self.__library:
                        self.__library.add_filename(loc)
            elif app.player.can_play_uri(uri):
                if uri not in self.__library:
                    self.__library.add([RemoteFile(uri)])
            else:
                error = True
                break
        Gtk.drag_finish(ctx, not error, False, etime)
        if error:
            ErrorMessage(
                self,
                _("Unable to add songs"),
                _("%s uses an unsupported protocol.") % util.bold(uri),
                escape_desc=False,
            ).run()
        else:
            if dirs:
                copool.add(
                    self.__library.scan, dirs, cofuncid="library", funcid="library"
                )

    def __songlist_key_press(self, songlist, event):
        return self.browser.key_pressed(event)

    def __songlist_drag_data_recv(self, view, *args):
        if self.browser.can_reorder:
            songs = view.get_songs()
            self.browser.reordered(songs)
        self.songlist.clear_sort()

    def __create_menu(self, player, library):
        def add_view_items(ag):
            act = Action(
                name="Information",
                label=_("_Information"),
                icon_name=Icons.DIALOG_INFORMATION,
            )
            act.connect("activate", self.__current_song_info)
            ag.add_action(act)

            act = Action(
                name="Jump", label=_("_Jump to Playing Song"), icon_name=Icons.GO_JUMP
            )
            self.__jump_to_current(True, None, True)
            act.connect("activate", self.__jump_to_current)
            ag.add_action_with_accel(act, "<Primary>J")

        def add_top_level_items(ag):
            ag.add_action(Action(name="File", label=_("_File")))
            ag.add_action(Action(name="Song", label=_("_Song")))
            ag.add_action(Action(name="View", label=_("_View")))
            ag.add_action(Action(name="Browse", label=_("_Browse")))
            ag.add_action(Action(name="Control", label=_("_Control")))
            ag.add_action(Action(name="Help", label=_("_Help")))

        ag = Gtk.ActionGroup.new("QuodLibetWindowActions")
        add_top_level_items(ag)
        add_view_items(ag)

        act = Action(
            name="AddFolders", label=_("_Add a Folder…"), icon_name=Icons.LIST_ADD
        )
        act.connect("activate", self.open_chooser)
        ag.add_action_with_accel(act, "<Primary>O")

        act = Action(name="AddFiles", label=_("_Add a File…"), icon_name=Icons.LIST_ADD)
        act.connect("activate", self.open_chooser)
        ag.add_action(act)

        act = Action(
            name="AddLocation", label=_("_Add a Location…"), icon_name=Icons.LIST_ADD
        )
        act.connect("activate", self.open_location)
        ag.add_action(act)

        act = Action(
            name="BrowseLibrary", label=_("Open _Browser"), icon_name=Icons.EDIT_FIND
        )
        ag.add_action(act)

        act = Action(
            name="Preferences",
            label=_("_Preferences"),
            icon_name=Icons.PREFERENCES_SYSTEM,
        )
        act.connect("activate", self.__preferences)
        ag.add_action(act)

        act = Action(name="Plugins", label=_("_Plugins"), icon_name=Icons.SYSTEM_RUN)
        act.connect("activate", self.__plugins)
        ag.add_action(act)

        act = Action(name="Quit", label=_("_Quit"), icon_name=Icons.APPLICATION_EXIT)
        act.connect("activate", lambda *x: self.destroy())
        ag.add_action_with_accel(act, "<Primary>Q")

        act = Action(
            name="EditTags", label=_("Edit _Tags"), icon_name=Icons.DOCUMENT_PROPERTIES
        )
        act.connect("activate", self.__current_song_prop)
        ag.add_action(act)

        act = Action(name="EditBookmarks", label=_("Edit Bookmarks…"))
        connect_obj(act, "activate", self.__edit_bookmarks, library.librarian, player)
        ag.add_action_with_accel(act, "<Primary>B")

        act = Action(
            name="Previous", label=_("Pre_vious"), icon_name=Icons.MEDIA_SKIP_BACKWARD
        )
        act.connect("activate", self.__previous_song)
        ag.add_action_with_accel(act, "<Primary>comma")

        act = Action(
            name="PlayPause", label=_("_Play"), icon_name=Icons.MEDIA_PLAYBACK_START
        )
        act.connect("activate", self.__play_pause)
        ag.add_action_with_accel(act, "<Primary>space")

        act = Action(name="Next", label=_("_Next"), icon_name=Icons.MEDIA_SKIP_FORWARD)
        act.connect("activate", self.__next_song)
        ag.add_action_with_accel(act, "<Primary>period")

        act = Action(name="Stop", label=_("Stop"), icon_name=Icons.MEDIA_PLAYBACK_STOP)
        act.connect("activate", self.__stop)
        ag.add_action(act)

        act = ToggleAction(name="StopAfter", label=_("Stop After This Song"))
        ag.add_action_with_accel(act, "<shift>space")

        # access point for the tray icon
        self.stop_after = act

        act = Action(name="Shortcuts", label=_("_Keyboard Shortcuts"))
        act.connect("activate", self.__keyboard_shortcuts)
        ag.add_action_with_accel(act, "<Primary>question")

        act = Action(name="About", label=_("_About"), icon_name=Icons.HELP_ABOUT)
        act.connect("activate", self.__show_about)
        ag.add_action_with_accel(act, None)

        act = Action(
            name="OnlineHelp", label=_("Online Help"), icon_name=Icons.HELP_BROWSER
        )

        def website_handler(*args):
            util.website(const.ONLINE_HELP)

        act.connect("activate", website_handler)
        ag.add_action_with_accel(act, "F1")

        act = Action(name="SearchHelp", label=_("Search Help"))

        def search_help_handler(*args):
            util.website(const.SEARCH_HELP)

        act.connect("activate", search_help_handler)
        ag.add_action_with_accel(act, None)

        act = Action(
            name="CheckUpdates",
            label=_("_Check for Updates…"),
            icon_name=Icons.NETWORK_SERVER,
        )

        def check_updates_handler(*args):
            d = UpdateDialog(self)
            d.run()
            d.destroy()

        act.connect("activate", check_updates_handler)
        ag.add_action_with_accel(act, None)

        act = Action(
            name="RefreshLibrary",
            label=_("_Scan Library"),
            icon_name=Icons.VIEW_REFRESH,
        )
        act.connect("activate", self.__rebuild, False)
        ag.add_action_with_accel(act, "<Primary>R")

        current = config.get("memory", "browser")
        try:
            browsers.get(current)
        except ValueError:
            current = browsers.name(browsers.default)

        first_action = None
        for Kind in browsers.browsers:
            name = browsers.name(Kind)
            index = browsers.index(name)
            action_name = "View" + Kind.__name__
            act = RadioAction(
                name=action_name, label=Kind.accelerated_name, value=index
            )
            act.join_group(first_action)
            first_action = first_action or act
            if name == current:
                act.set_active(True)
            ag.add_action_with_accel(act, "<Primary>%d" % ((index + 1) % 10,))
        assert first_action
        self._browser_action = first_action

        def action_callback(view_action, current_action):
            current = browsers.name(browsers.get(current_action.get_current_value()))
            self._select_browser(view_action, current, library, player)

        first_action.connect("changed", action_callback)

        for Kind in browsers.browsers:
            action = "Browser" + Kind.__name__
            label = Kind.accelerated_name
            name = browsers.name(Kind)
            index = browsers.index(name)
            act = Action(name=action, label=label)

            def browser_activate(action, browser_cls):
                LibraryBrowser.open(browser_cls, library, player)

            act.connect("activate", browser_activate, Kind)
            ag.add_action_with_accel(act, "<Primary><alt>%d" % ((index + 1) % 10,))

        ui = Gtk.UIManager()
        ui.insert_action_group(ag, -1)

        menustr = MENU % {
            "views": browser_menu_items(),
            "browsers": secondary_browser_menu_items(),
            "filters_menu": FilterMenu.MENU,
        }
        ui.add_ui_from_string(menustr)
        self._filter_menu = FilterMenu(library, player, ui)

        # Cute. So. UIManager lets you attach tooltips, but when they're
        # for menu items, they just get ignored. So here I get to actually
        # attach them.
        ui.get_widget("/Menu/File/RefreshLibrary").set_tooltip_text(
            _("Check for changes in your library")
        )

        return ui

    def __show_about(self, *args):
        about = AboutDialog(self, app)
        about.run()
        about.destroy()

    def select_browser(self, browser_key, library, player):
        """Given a browser name (see browsers.get()) changes the current
        browser.

        Returns True if the passed browser ID is known and the change
        was initiated.
        """

        try:
            Browser = browsers.get(browser_key)
        except ValueError:
            return False

        action_name = f"View{Browser.__name__}"
        for action in self._browser_action.get_group():
            if action.get_name() == action_name:
                action.set_active(True)
                return True
        return False

    def _select_browser(self, activator, current, library, player, restore=False):
        Browser = browsers.get(current)

        window = self.get_window()
        if window:
            window.set_cursor(Gdk.Cursor.new(Gdk.CursorType.WATCH))

        # Wait for the cursor to update before continuing
        while Gtk.events_pending():
            Gtk.main_iteration()

        config.set("memory", "browser", current)
        if self.browser:
            if not (self.browser.uses_main_library and Browser.uses_main_library):
                self.songlist.clear()
            container = self.browser.__container
            self.browser.unpack(container, self.songpane)
            if self.browser.accelerators:
                self.remove_accel_group(self.browser.accelerators)
            container.destroy()
            self.browser.destroy()
        self.browser = Browser(library)
        self.browser.connect("songs-selected", self.__browser_cb, library, player)
        self.browser.connect("songs-activated", self.__browser_activate)
        if restore:
            self.browser.restore()
            self.browser.activate()
        self.browser.finalize(restore)
        if not restore:
            self.browser.unfilter()
        if self.browser.can_reorder:
            self.songlist.enable_drop()
        elif self.browser.dropped:
            self.songlist.enable_drop(False)
        else:
            self.songlist.disable_drop()
        if self.browser.accelerators:
            self.add_accel_group(self.browser.accelerators)
        self.set_sortability()
        container = self.browser.__container = self.browser.pack(self.songpane)

        # Reset the cursor when done loading the browser
        if window:
            GLib.idle_add(window.set_cursor, None)

        player.replaygain_profiles[1] = self.browser.replaygain_profiles
        player.reset_replaygain()
        self.__browserbox.add(container)
        container.show()
        self._filter_menu.set_browser(self.browser)
        self.__hide_headers()

    def set_sortability(self):
        self.songlist.sortable = not self.browser.can_reorder

    def __update_paused(self, player, paused):
        menu = self.ui.get_widget("/Menu/Control/PlayPause")
        image = menu.get_image()

        if paused:
            label, icon = _("_Play"), Icons.MEDIA_PLAYBACK_START
        else:
            label, icon = _("P_ause"), Icons.MEDIA_PLAYBACK_PAUSE

        menu.set_label(label)
        image.set_from_icon_name(icon, Gtk.IconSize.MENU)

    def __song_ended(self, player, song, stopped):
        # Check if the song should be removed, based on the
        # active filter of the current browser.
        active_filter = self.browser.active_filter
        if song and active_filter and not active_filter(song):
            iter_ = self.songlist.model.find(song)
            if iter_:
                self.songlist.remove_iters([iter_])

        if self.stop_after.get_active():
            player.paused = True
            self.stop_after.set_active(False)

    def __song_changed(self, library, songs, player):
        if player.info in songs:
            self.__update_title(player)

    def __update_title(self, player):
        song = player.info
        title = "Quod Libet"
        if song:
            tag = config.gettext("settings", "window_title_pattern")
            if tag:
                title = song.comma(tag) + " - " + title
        self.set_title(title)

    def __song_started(self, player, song):
        self.__update_title(player)

        for wid in [
            "Control/Next",
            "Control/StopAfter",
            "Song/EditTags",
            "Song/Information",
            "Song/EditBookmarks",
            "Song/Jump",
        ]:
            self.ui.get_widget("/Menu/" + wid).set_sensitive(bool(song))

        # don't jump on stream changes (player.info != player.song)
        main_should_jump = (
            song
            and player.song is song
            and not self.songlist._activated
            and config.getboolean("settings", "jump")
            and self.songlist.sourced
        )
        queue_should_jump = (
            song
            and player.song is song
            and not self.qexpander.queue._activated
            and config.getboolean("settings", "jump")
            and self.qexpander.queue.sourced
            and config.getboolean("memory", "queue_keep_songs")
        )
        if main_should_jump:
            self.__jump_to_current(False, self.songlist)
        elif queue_should_jump:
            self.__jump_to_current(False, self.qexpander.queue)

    def __play_pause(self, *args):
        app.player.playpause()

    def __stop(self, *args):
        app.player.stop()

    def __jump_to_current(self, explicit, songlist=None, force_scroll=False):
        """Select/scroll to the current playing song in the playlist.
        If it can't be found tell the browser to properly fill the playlist
        with an appropriate selection containing the song.

        explicit means that the jump request comes from the user and not
        from an event like song-started.

        songlist is the songlist to be jumped within. Usually the main song
        list or the queue. If None, the currently sourced songlist will be
        used.

        force_scroll will ask the browser to refill the playlist in any case.
        """

        def idle_jump_to(song, select):
            ok = songlist.jump_to_song(song, select=select)
            if ok:
                songlist.grab_focus()
            return False

        if not songlist:
            if (
                config.getboolean("memory", "queue_keep_songs")
                and self.qexpander.queue.sourced
            ):
                songlist = self.qexpander.queue
            else:
                songlist = self.songlist

        if app.player is None:
            return

        song = app.player.song

        # We are not playing a song
        if song is None:
            return

        if not force_scroll:
            ok = songlist.jump_to_song(song, select=explicit)
        else:
            assert explicit
            ok = False

        if ok:
            songlist.grab_focus()
        elif explicit:
            # if we can't find it and the user requested it, try harder
            self.browser.scroll(song)
            # We need to wait until the browser has finished
            # scrolling/filling and the songlist is ready.
            # Not perfect, but works for now.
            GLib.idle_add(idle_jump_to, song, explicit, priority=GLib.PRIORITY_LOW)

    def __next_song(self, *args):
        app.player.next()

    def __previous_song(self, *args):
        app.player.previous()

    def __rebuild(self, activator, force):
        scan_library(self.__library, force)

    # Set up the preferences window.
    def __preferences(self, activator):
        window = PreferencesWindow(self)
        window.show()

    def __plugins(self, activator):
        window = PluginWindow(self)
        window.show()

    def open_location(self, action):
        name = GetStringDialog(
            self,
            _("Add a Location"),
            _("Enter the location of an audio file:"),
            button_label=_("_Add"),
            button_icon=Icons.LIST_ADD,
        ).run()
        if name:
            if not uri_is_valid(name):
                ErrorMessage(
                    self,
                    _("Unable to add location"),
                    _("%s is not a valid location.") % util.bold(name),
                ).run()
            elif not app.player.can_play_uri(name):
                ErrorMessage(
                    self,
                    _("Unable to add location"),
                    _("%s uses an unsupported protocol.") % (util.bold(name)),
                    escape_desc=False,
                ).run()
            else:
                if name not in self.__library:
                    self.__library.add([RemoteFile(name)])

    def open_chooser(self, action):
        if action.get_name() == "AddFolders":
            fns = choose_folders(self, _("Add Music"), _("_Add Folders"))
            if fns:
                # scan them
                copool.add(
                    self.__library.scan, fns, cofuncid="library", funcid="library"
                )
        else:
            patterns = ["*" + path2fsn(k) for k in formats.loaders.keys()]
            choose_filter = create_chooser_filter(_("Music Files"), patterns)
            fns = choose_files(self, _("Add Music"), _("_Add Files"), choose_filter)
            if fns:
                for filename in fns:
                    self.__library.add_filename(filename)

    def __songs_popup_menu(self, songlist):
        path, col = songlist.get_cursor()
        header = col.header_name
        menu = self.songlist.menu(header, self.browser, self.__library)
        if menu is not None:
            return self.songlist.popup_menu(menu, 0, Gtk.get_current_event_time())
        return None

    def __current_song_prop(self, *args):
        song = app.player.song
        if song:
            librarian = self.__library.librarian
            window = SongProperties(librarian, [song], parent=self)
            window.show()

    def __current_song_info(self, *args):
        song = app.player.song
        if song:
            librarian = self.__library.librarian
            window = Information(librarian, [song], self)
            window.show()

    def __browser_activate(self, browser):
        app.player._reset()

    def __browser_cb(self, browser, songs, sorted, library, player):
        if browser.background:
            bg = background_filter()
            if bg:
                songs = list(filter(bg, songs))
        self.songlist.set_songs(songs, sorted)

        # After the first time the browser activates, which should always
        # happen if we start up and restore, restore the playing song.
        # Because the browser has send us songs we can be sure it has
        # registered all its libraries.
        if self.__first_browser_set:
            self.__first_browser_set = False

            song = library.librarian.get(config.get("memory", "song"))
            seek_pos = config.getfloat("memory", "seek", 0)
            config.set("memory", "seek", 0)
            if song is not None:
                player.setup(self.playlist, song, seek_pos, False)

            if self.__restore_cb:
                self.__restore_cb()
                self.__restore_cb = None

    def __hide_headers(self, activator=None):
        for column in self.songlist.get_columns():
            if self.browser.headers is None:
                column.set_visible(True)
            else:
                for tag in util.tagsplit(column.header_name):
                    if tag in self.browser.headers:
                        column.set_visible(True)
                        break
                else:
                    column.set_visible(False)

    def __cols_changed(self, songlist):
        headers = [col.header_name for col in songlist.get_columns()]
        try:
            headers.remove("~current")
        except ValueError:
            pass
        if len(headers) == len(get_columns()):
            # Not an addition or removal (handled separately)
            set_columns(headers)
            SongList.headers = headers

    def __make_query(self, query):
        if self.browser.can_filter_text():
            self.browser.filter_text(query.encode("utf-8"))
            self.browser.activate()

    def __set_totals(self, info, songs):
        length = sum(song.get("~#length", 0) for song in songs)
        t = self.browser.status_text(
            count=len(songs), time=util.format_time_preferred(length)
        )
        self.statusbar.set_default_text(t)


class ConfirmEnqueue(qltk.Message):
    def __init__(self, parent, count):
        title = (
            ngettext(
                "Are you sure you want to enqueue %d song?",
                "Are you sure you want to enqueue %d songs?",
                count,
            )
            % count
        )
        description = ""

        super().__init__(
            Gtk.MessageType.WARNING, parent, title, description, Gtk.ButtonsType.NONE
        )

        self.add_button(_("_Cancel"), Gtk.ResponseType.CANCEL)
        self.add_icon_button(_("_Enqueue"), Icons.LIST_ADD, Gtk.ResponseType.YES)
